#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from typing import Any, Callable, Dict, Optional, Union

from merlin.models.tf.blocks.core.aggregation import CosineSimilarity
from merlin.models.tf.blocks.core.transformations import RenameFeatures
from merlin.models.tf.blocks.retrieval.base import DualEncoderBlock
from merlin.models.tf.features.embedding import EmbeddingFeatures, EmbeddingOptions
from merlin.schema import Schema, Tags

LOG = logging.getLogger("merlin_models")


class QueryItemIdsEmbeddingsBlock(DualEncoderBlock):
    """
    An encoder for user ids and item ids

    Parameters
    ----------
    schema : Schema
        The `Schema` with the input features
    dim : int
        Dimension of the user and item embeddings
    query_id_tag : _type_, optional
        The tag to select the user id feature, by default `Tags.USER_ID`
    item_id_tag : _type_, optional
        The tag to select the item id feature, by default `Tags.ITEM_ID`
    embeddings_initializers : Optional[Dict[str, Callable[[Any], None]]], optional
        Dict where keys are feature names and values are callable to initialize embedding tables
    """

    def __init__(
        self,
        schema: Schema,
        dim: int,
        query_id_tag=Tags.USER_ID,
        item_id_tag=Tags.ITEM_ID,
        embeddings_initializers: Optional[Dict[str, Callable[[Any], None]]] = None,
        **kwargs,
    ):
        query_schema = schema.select_by_tag(query_id_tag)
        item_schema = schema.select_by_tag(item_id_tag)
        embedding_options = EmbeddingOptions(
            embedding_dim_default=dim, embeddings_initializers=embeddings_initializers
        )

        rename_features = RenameFeatures(
            {query_id_tag: "query", item_id_tag: "item"}, schema=schema
        )
        post = kwargs.pop("post", None)
        if post:
            post = rename_features.connect(post)
        else:
            post = rename_features

        embedding = lambda s: EmbeddingFeatures.from_schema(  # noqa
            s, embedding_options=embedding_options, aggregation="concat"
        )

        super().__init__(  # type: ignore
            embedding(query_schema), embedding(item_schema), post=post, **kwargs
        )

    def export_embedding_table(self, table_name: Union[str, Tags], export_path: str, gpu=True):
        if table_name in ("item", Tags.ITEM_ID):
            return self.item_block().block.export_embedding_table(table_name, export_path, gpu=gpu)

        return self.query_block().block.export_embedding_table(table_name, export_path, gpu=gpu)

    def embedding_table_df(self, table_name: Union[str, Tags], gpu=True):
        if table_name in ("item", Tags.ITEM_ID):
            return self.item_block().block.embedding_table_df(table_name, gpu=gpu)

        return self.query_block().block.embedding_table_df(table_name, gpu=gpu)


def MatrixFactorizationBlock(
    schema: Schema,
    dim: int,
    query_id_tag=Tags.USER_ID,
    item_id_tag=Tags.ITEM_ID,
    embeddings_initializers: Optional[Dict[str, Callable[[Any], None]]] = None,
    aggregation=CosineSimilarity(),
    **kwargs,
):
    """
    Returns a block for Matrix Factorization, which created the user and
    item embeddings based on the `schema` and computes the dot product
    between user and item L2-norm embeddings

    Parameters
    ----------
    schema : Schema
        The `Schema` with the input features
    dim : int
        Dimension of the user and item embeddings
    query_id_tag : _type_, optional
        The tag to select the user id feature, by default `Tags.USER_ID`
    item_id_tag : _type_, optional
        The tag to select the item id feature, by default `Tags.ITEM_ID`
    embeddings_initializers : Optional[Dict[str, Callable[[Any], None]]], optional
        Dict where keys are feature names and values are callable to initialize
        embedding tables
    aggregation : _type_, optional
        Aggregation of the user and item embeddings, by default CosineSimilarity()

    Returns
    -------
    QueryItemIdsEmbeddingsBlock
        A block that encodes user ids and item ids into embeddings and computes their
        dot product
    """
    return QueryItemIdsEmbeddingsBlock(
        schema=schema,
        dim=dim,
        query_id_tag=query_id_tag,
        item_id_tag=item_id_tag,
        embeddings_initializers=embeddings_initializers,
        aggregation=aggregation,
        **kwargs,
    )



from collections import Sequence as SequenceCollection
from typing import Dict, List, Optional, Protocol, Union, runtime_checkable

import tensorflow as tf

import merlin.io
from merlin.models.tf.blocks.core.base import Block, BlockContext
from merlin.models.tf.blocks.core.combinators import SequentialBlock
from merlin.models.tf.metrics.ranking import RankingMetric
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.models.tf.typing import TabularData
from merlin.models.tf.utils.mixins import LossMixin, MetricsMixin, ModelLikeBlock
from merlin.models.utils.dataset import unique_rows_by_features
from merlin.schema import Schema, Tags


class MetricsComputeCallback(tf.keras.callbacks.Callback):
    def __init__(self, train_metrics_steps=1, **kwargs):
        self.train_metrics_steps = train_metrics_steps
        self._is_fitting = False
        self._is_first_batch = True
        super().__init__(**kwargs)

    def on_train_begin(self, logs=None):
        self._is_fitting = True

    def on_train_end(self, logs=None):
        self._is_fitting = False

    def on_epoch_begin(self, epoch, logs=None):
        self._is_first_batch = True

    def on_train_batch_begin(self, batch, logs=None):
        value = self.train_metrics_steps > 0 and (
            self._is_first_batch or batch % self.train_metrics_steps == 0
        )
        self.model._should_compute_train_metrics_for_batch.assign(value)

    def on_train_batch_end(self, batch, logs=None):
        self._is_first_batch = False

    def on_test_begin(self, logs=None):
        self.model._should_compute_eval_metrics_as_in_training.assign(self._is_fitting)

    def on_test_end(self, logs=None):
        self.model._should_compute_eval_metrics_as_in_training.assign(False)


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class ModelBlock(Block, tf.keras.Model):
    def __init__(self, block: Block, **kwargs):
        super().__init__(**kwargs)
        self.block = block

    def call(self, inputs, **kwargs):
        outputs = self.block(inputs, **kwargs)
        return outputs

    def build(self, input_shapes):
        self.block.build(input_shapes)

        if not hasattr(self.build, "_is_default"):
            self._build_input_shape = input_shapes
        self.built = True

    def compute_output_shape(self, input_shape):
        return self.block.compute_output_shape(input_shape)

    @property
    def schema(self) -> Schema:
        return self.block.schema

    @classmethod
    def from_config(cls, config, custom_objects=None):
        block = tf.keras.utils.deserialize_keras_object(config.pop("block"))

        return cls(block, **config)

    def get_config(self):
        return {"block": tf.keras.utils.serialize_keras_object(self.block)}


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class Model(tf.keras.Model, LossMixin, MetricsMixin):
    def __init__(
        self,
        *blocks: Union[Block, ModelLikeBlock],
        context: Optional[BlockContext] = None,
        **kwargs,
    ):
        super(Model, self).__init__(**kwargs)
        context = context or BlockContext()
        if (
            len(blocks) == 1
            and isinstance(blocks[0], SequentialBlock)
            and isinstance(blocks[0].layers[-1], ModelLikeBlock)
        ):
            self.block = blocks[0]
        else:
            if not isinstance(blocks[-1], ModelLikeBlock):
                raise ValueError("Last block must be able to calculate loss & metrics.")
            self.block = SequentialBlock(blocks, context=context)
        if not getattr(self.block, "_context", None):
            self.block._set_context(context)
        self.context = context
        self._is_fitting = False

        # Initializing model control flags controlled by MetricsComputeCallback()
        self._should_compute_train_metrics_for_batch = tf.Variable(
            dtype=tf.bool,
            name="should_compute_train_metrics_for_batch",
            trainable=False,
            synchronization=tf.VariableSynchronization.NONE,
            initial_value=lambda: False,
        )
        self._should_compute_eval_metrics_as_in_training = tf.Variable(
            dtype=tf.bool,
            name="should_compute_eval_metrics_as_in_training",
            trainable=False,
            synchronization=tf.VariableSynchronization.NONE,
            initial_value=lambda: False,
        )

    def call(self, inputs, **kwargs):
        outputs = self.block(inputs, **kwargs)
        return outputs

    # @property
    # def inputs(self):
    #     return self.block.inputs

    @property
    def first(self):
        return self.block.layers[0]

    @property
    def last(self):
        return self.block.layers[-1]

    @property
    def loss_block(self) -> ModelLikeBlock:
        return self.block.last if isinstance(self.block, SequentialBlock) else self.block

    @property
    def schema(self) -> Schema:
        return self.block.schema

    @classmethod
    def from_block(
        cls,
        block: Block,
        schema: Schema,
        input_block: Optional[Block] = None,
        prediction_tasks: Optional[
            Union["PredictionTask", List["PredictionTask"], "ParallelPredictionBlock"]
        ] = None,
        **kwargs,
    ) -> "Model":
        """Create a model from a `block`
        Parameters
        ----------
        block: Block
            The block to wrap in-between an InputBlock and prediction task(s)
        schema: Schema
            Schema to use for the model.
        input_block: Optional[Block]
            Block to use as input.
        prediction_tasks: Optional[
            Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
        ]
            Prediction tasks to use.
        """

        return block.to_model(
            schema, input_block=input_block, prediction_tasks=prediction_tasks, **kwargs
        )

    def compute_loss(
        self,
        inputs: Union[tf.Tensor, TabularData],
        targets: Union[tf.Tensor, TabularData],
        compute_metrics=False,
        training: bool = False,
        **kwargs,
    ) -> tf.Tensor:
        return self.loss_block.compute_loss(
            inputs, targets, training=training, compute_metrics=compute_metrics, **kwargs
        )

    def calculate_metrics(
        self,
        outputs,
        mode: str = "val",
        forward: bool = True,
        training: bool = False,
        **kwargs,
    ) -> Dict[str, Union[Dict[str, tf.Tensor], tf.Tensor]]:
        return self.loss_block.calculate_metrics(
            outputs=outputs, mode=mode, forward=forward, training=training, **kwargs
        )

    def metric_results(self, mode=None):
        return self.loss_block.metric_results(mode=mode)

    def train_step(self, inputs):
        """Custom train step using the `compute_loss` method."""

        with tf.GradientTape() as tape:
            if isinstance(inputs, tuple):
                if len(inputs) == 1:
                    inputs = inputs[0]
                    targets = None
                else:
                    inputs, targets = inputs
            else:
                targets = None

            predictions = self(inputs, training=True)
            loss = self.compute_loss(
                predictions,
                targets,
                training=True,
                compute_metrics=self._should_compute_train_metrics_for_batch,
            )
            tf.assert_rank(
                loss,
                0,
                "The loss tensor should have rank 0. "
                "Check if you are using a tf.keras.losses.Loss with 'reduction' "
                "properly set",
            )
            assert loss.dtype == tf.float32, (
                f"The loss dtype should be tf.float32 but is rather {loss.dtype}. "
                "Ensure that your model output has tf.float32 dtype, as "
                "that should be the case when using mixed_float16 policy "
                "to avoid numerical instabilities."
            )

            regularization_loss = tf.reduce_sum(self.losses)

            total_loss = tf.add_n([loss, regularization_loss])

            if getattr(self.optimizer, "get_scaled_loss", False):
                scaled_loss = self.optimizer.get_scaled_loss(total_loss)

        # If mixed precision (mixed_float16 policy) is enabled
        # (and the optimizer is automatically wrapped by
        #  tensorflow.keras.mixed_precision.LossScaleOptimizer())
        if getattr(self.optimizer, "get_scaled_loss", False):
            scaled_gradients = tape.gradient(scaled_loss, self.trainable_variables)
            gradients = self.optimizer.get_unscaled_gradients(scaled_gradients)
        else:
            gradients = tape.gradient(total_loss, self.trainable_variables)

        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))

        metrics = self.loss_block.metric_result_dict()

        metrics["loss"] = loss
        metrics["regularization_loss"] = regularization_loss
        metrics["total_loss"] = total_loss

        return metrics

    def test_step(self, inputs):
        """Custom test step using the `compute_loss` method."""

        if isinstance(inputs, tuple):
            if len(inputs) == 1:
                inputs = inputs[0]
                targets = None
            else:
                inputs, targets = inputs
        else:
            targets = None

        def compute_loss_metrics(training):
            predictions = self(inputs, training=training)
            loss = self.compute_loss(predictions, targets, training=training, compute_metrics=True)
            tf.assert_rank(
                loss,
                0,
                "The loss tensor should have rank 0. "
                "Check if you are using a tf.keras.losses.Loss with 'reduction' "
                "properly set",
            )
            return loss

        loss = tf.cond(
            self._should_compute_eval_metrics_as_in_training,
            lambda: compute_loss_metrics(training=True),
            lambda: compute_loss_metrics(training=False),
        )

        # Casting regularization loss to fp16 if needed to match the main loss
        regularization_loss = tf.cast(tf.reduce_sum(self.losses), loss.dtype)

        total_loss = loss + regularization_loss

        metrics = self.loss_block.metric_result_dict()
        metrics["loss"] = loss
        metrics["regularization_loss"] = regularization_loss
        metrics["total_loss"] = total_loss

        return metrics

    def fit(
        self,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        verbose="auto",
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        class_weight=None,
        sample_weight=None,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_batch_size=None,
        validation_freq=1,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
        train_metrics_steps=1,
        **kwargs,
    ):
        # Check if merlin-dataset is passed
        if hasattr(x, "to_ddf"):
            if not batch_size:
                raise ValueError("batch_size must be specified when using merlin-dataset.")
            from merlin.models.tf.dataset import BatchedDataset

            x = BatchedDataset(x, batch_size=batch_size, **kwargs)

        if hasattr(validation_data, "to_ddf"):
            if not batch_size:
                raise ValueError("batch_size must be specified when using merlin-dataset.")
            from merlin.models.tf.dataset import BatchedDataset

            kwargs.pop("shuffle", None)
            validation_data = BatchedDataset(
                validation_data, batch_size=batch_size, shuffle=False, **kwargs
            )

        callbacks = self._add_metrics_callback(callbacks, train_metrics_steps)

        return super().fit(
            x,
            y,
            batch_size,
            epochs,
            verbose,
            callbacks,
            validation_split,
            validation_data,
            shuffle,
            class_weight,
            sample_weight,
            initial_epoch,
            steps_per_epoch,
            validation_steps,
            validation_batch_size,
            validation_freq,
            max_queue_size,
            workers,
            use_multiprocessing,
        )

    def _add_metrics_callback(self, callbacks, train_metrics_steps):
        if callbacks is None:
            callbacks = []

        if isinstance(callbacks, SequenceCollection):
            callbacks = list(callbacks)
        else:
            callbacks = [callbacks]

        callback_types = [type(callback) for callback in callbacks]
        if MetricsComputeCallback not in callback_types:
            # Adding a callback to control metrics computation
            callbacks.append(MetricsComputeCallback(train_metrics_steps))

        return callbacks

    def evaluate(
        self,
        x=None,
        y=None,
        batch_size=None,
        verbose=1,
        sample_weight=None,
        steps=None,
        callbacks=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
        return_dict=False,
        **kwargs,
    ):
        # Check if merlin-dataset is passed
        if hasattr(x, "to_ddf"):
            if not batch_size:
                raise ValueError("batch_size must be specified when using merlin-dataset.")
            from merlin.models.tf.dataset import BatchedDataset

            x = BatchedDataset(x, batch_size=batch_size, shuffle=False, **kwargs)

        # Load BruteForce top-k evaluation
        # using evaluation_candidates as indices.
        if isinstance(self, RetrievalModel):
            self.check_for_retrieval_task()
            # We need to load the top-k indices at each call of the evaluate()
            # to ensure most updated pre-computed embeddings are loaded.
            self = self._load_or_update_candidates_index()

        return super().evaluate(
            x,
            y,
            batch_size,
            verbose,
            sample_weight,
            steps,
            callbacks,
            max_queue_size,
            workers,
            use_multiprocessing,
            return_dict,
            **kwargs,
        )

    def batch_predict(
        self, dataset: merlin.io.Dataset, batch_size: int, **kwargs
    ) -> merlin.io.Dataset:
        """Batched prediction using the Dask.
        Parameters
        ----------
        dataset: merlin.io.Dataset
            Dataset to predict on.
        batch_size: int
            Batch size to use for prediction.
        Returns merlin.io.Dataset
        -------
        """
        if hasattr(dataset, "schema"):
            if not set(self.schema.column_names).issubset(set(dataset.schema.column_names)):
                raise ValueError(
                    f"Model schema {self.schema.column_names} does not match dataset schema"
                    + f" {dataset.schema.column_names}"
                )

        # Check if merlin-dataset is passed
        if hasattr(dataset, "to_ddf"):
            dataset = dataset.to_ddf()

        from merlin.models.tf.utils.batch_utils import TFModelEncode

        model_encode = TFModelEncode(self, batch_size=batch_size, **kwargs)
        predictions = dataset.map_partitions(model_encode)

        return merlin.io.Dataset(predictions)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        block = tf.keras.utils.deserialize_keras_object(config.pop("block"))

        return cls(block, **config)

    def get_config(self):
        return {"block": tf.keras.utils.serialize_keras_object(self.block)}


@runtime_checkable
class RetrievalBlock(Protocol):
    def query_block(self) -> Block:
        ...

    def item_block(self) -> Block:
        ...


class RetrievalModel(Model):
    """Embedding-based retrieval model."""

    def __init__(
        self,
        *blocks: Union[Block, ModelLikeBlock],
        context: Optional[BlockContext] = None,
        evaluation_candidates: merlin.io.Dataset = None,
        **kwargs,
    ):
        super().__init__(*blocks, context=context, **kwargs)

        if not any(isinstance(b, RetrievalBlock) for b in self.block):
            raise ValueError("Model must contain a `RetrievalBlock`.")

        self.evaluation_candidates = evaluation_candidates

    @property
    def retrieval_block(self) -> RetrievalBlock:
        return next(b for b in self.block if isinstance(b, RetrievalBlock))

    def query_embeddings(
        self,
        dataset: merlin.io.Dataset,
        query_tag=Tags.USER,
        query_id_tag=Tags.USER_ID,
        batch_size=None,
    ) -> merlin.io.Dataset:
        """Export query embeddings from the model.
        Parameters
        ----------
        dataset: merlin.io.Dataset
            Dataset to export embeddings from.
        query_tag: str
            Tag to use for the query.
        query_id_tag: str
            Tag to use for the query id.
        batch_size: int
            Batch size to use for embedding extraction.
        Returns
        -------
        merlin.io.Dataset
        """
        from merlin.models.tf.utils.batch_utils import QueryEmbeddings

        get_user_emb = QueryEmbeddings(self, batch_size=batch_size)

        dataset = unique_rows_by_features(dataset, query_tag, query_id_tag).to_ddf()
        embeddings = dataset.map_partitions(get_user_emb)

        return merlin.io.Dataset(embeddings)

    def item_embeddings(
        self,
        dataset: merlin.io.Dataset,
        item_tag=Tags.ITEM,
        item_id_tag=Tags.ITEM_ID,
        batch_size=None,
    ) -> merlin.io.Dataset:
        """Export item embeddings from the model.
        Parameters
        ----------
        dataset: merlin.io.Dataset
            Dataset to export embeddings from.
        item_tag: str
            Tag to use for the item.
        item_id_tag: str
            Tag to use for the item id.
        batch_size: int
            Batch size to use for embedding extraction.
        Returns
        -------
        merlin.io.Dataset
        """
        from merlin.models.tf.utils.batch_utils import ItemEmbeddings

        get_item_emb = ItemEmbeddings(self, batch_size=batch_size)

        dataset = unique_rows_by_features(dataset, item_tag, item_id_tag).to_ddf()
        embeddings = dataset.map_partitions(get_item_emb)

        return merlin.io.Dataset(embeddings)

    def check_for_retrieval_task(self):
        if not (
            getattr(self, "loss_block", None)
            and getattr(self.loss_block, "set_retrieval_cache_query", None)
        ):
            raise ValueError(
                "Your retrieval model should contain an ItemRetrievalTask "
                "in the end (loss_block)."
            )

    def set_retrieval_candidates_for_evaluation(self, candidates: merlin.io.Dataset):
        self.evaluation_candidates = unique_rows_by_features(candidates, Tags.ITEM, Tags.ITEM_ID)

        ranking_metrics = list(
            [metric for metric in self.loss_block.eval_metrics if isinstance(metric, RankingMetric)]
        )
        self._k = tf.reduce_max([metric.k for metric in ranking_metrics])

    def _load_or_update_candidates_index(self, **kwargs):
        """Update the model with a top-k evaluation block."""
        import merlin.models.tf as ml

        self.check_for_retrieval_task()
        if self.evaluation_candidates is None:
            raise ValueError(
                "You need to specify the set of negatives to use for evaluation "
                "via `set_retrieval_candidates_for_evaluation` method"
            )

        if self.loss_block.pre_eval_topk is None:
            topk_index = ml.TopKIndexBlock.from_block(
                self.retrieval_block.item_block(),
                data=self.evaluation_candidates,
                k=self._k,
                context=self.context,
                **kwargs,
            )
            self.loss_block.pre_eval_topk = topk_index  # type: ignore
        else:
            if not self._should_compute_eval_metrics_as_in_training:
                self.loss_block.pre_eval_topk.update_from_block(
                    self.retrieval_block.item_block(), self.evaluation_candidates
                )

        # set cache_query to True in the ItemRetrievalScorer
        self.loss_block.set_retrieval_cache_query(True)  # type: ignore

        return self

    def to_top_k_recommender(self, data: merlin.io.Dataset, k: int, **kwargs) -> ModelBlock:
        """Convert the model to a Top-k Recommender.
        Parameters
        ----------
        data: merlin.io.Dataset
            Dataset to convert to a Top-k Recommender.
        k: int
            Number of recommendations to make.
        Returns
        -------
        SequentialBlock
        """
        import merlin.models.tf as ml

        topk_index = ml.TopKIndexBlock.from_block(
            self.retrieval_block.item_block(), data=data, k=k, **kwargs
        )
        recommender = self.retrieval_block.query_block().connect(topk_index)

        return ModelBlock(recommender)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union

import numpy as np
import tensorflow as tf
from packaging import version

from merlin.models.tf.typing import TabularData

if version.parse(tf.__version__) < version.parse("2.3.0"):
    try:
        from tfdlpack import from_dlpack
    except ModuleNotFoundError as e:
        message = "If using TensorFlow < 2.3.0, you must install tfdlpack-gpu extension library"
        raise ModuleNotFoundError(message) from e

else:
    from tensorflow.experimental.dlpack import from_dlpack


def get_output_sizes_from_schema(schema, batch_size=0, max_sequence_length=None):
    sizes = {}
    for feature in schema:
        name = feature.name
        if feature.is_list:
            sizes[name] = tf.TensorShape(
                [
                    batch_size,
                    max_sequence_length if max_sequence_length else feature.value_count.max,
                ]
            )
        elif feature.HasField("shape"):
            sizes[name] = tf.TensorShape([batch_size] + [d.size for d in feature.shape.dim])
        else:
            sizes[name] = tf.TensorShape([batch_size, 1])

    return sizes


def calculate_batch_size_from_input_shapes(input_shapes):
    values = []

    for val in input_shapes.values():
        if isinstance(val, tuple) and isinstance(val[0], tf.TensorShape):
            values.append(val[1])
        else:
            values.append(val)

    return values[0][0]


def maybe_serialize_keras_objects(
    self,
    config,
    maybe_serialize_keys,
):
    for key in maybe_serialize_keys:
        maybe_value = getattr(self, key, None)
        if maybe_value:
            if isinstance(maybe_value, dict):
                config[key] = {
                    k: tf.keras.utils.serialize_keras_object(v) for k, v in maybe_value.items()
                }
            elif isinstance(maybe_value, (list, tuple)):
                config[key] = [tf.keras.utils.serialize_keras_object(v) for v in maybe_value]
            else:
                config[key] = tf.keras.utils.serialize_keras_object(maybe_value)

    return config


def maybe_deserialize_keras_objects(
    config, to_deserialize, deserialize_fn=tf.keras.utils.deserialize_keras_object
):
    if isinstance(to_deserialize, list):
        to_deserialize = {k: deserialize_fn for k in to_deserialize}

    custom_objects = {}

    for key, fn in to_deserialize.items():
        maybe_val = config.get(key, None)
        if maybe_val:
            if isinstance(maybe_val, list):
                config[key] = [fn(v, custom_objects=custom_objects) for v in maybe_val]
            else:
                config[key] = fn(maybe_val, custom_objects=custom_objects)

    return config


def rescore_false_negatives(
    positive_item_ids: tf.Tensor,
    neg_samples_item_ids: tf.Tensor,
    negative_scores: tf.Tensor,
    false_negatives_score: float,
):
    """
    Zeroes the logits of accidental negatives.
    """
    # Removing dimensions of size 1 from the shape of the item ids, if applicable
    positive_item_ids = tf.cast(tf.squeeze(positive_item_ids), neg_samples_item_ids.dtype)
    neg_samples_item_ids = tf.squeeze(neg_samples_item_ids)

    # Reshapes positive and negative ids so that false_negatives_mask matches the scores shape
    false_negatives_mask = tf.equal(
        tf.expand_dims(positive_item_ids, -1), tf.expand_dims(neg_samples_item_ids, 0)
    )

    # Setting a very small value for false negatives (accidental hits) so that it has
    # negligicle effect on the loss functions
    negative_scores = tf.where(
        false_negatives_mask,
        tf.ones_like(negative_scores) * false_negatives_score,
        negative_scores,
    )

    valid_negatives_mask = tf.logical_not(false_negatives_mask)

    return tf.squeeze(negative_scores), valid_negatives_mask


def extract_topk(k, predictions, labels):
    # Computes the number of relevant items per row (before extracting only the top-k)
    label_relevant_counts = tf.reduce_sum(labels, axis=-1)
    # Limits k to the number of prediction scores
    k = tf.minimum(k, tf.shape(predictions)[-1])
    topk_predictions, topk_indices = tf.math.top_k(predictions, k)
    topk_labels = gather_torch_like(labels, topk_indices, k)
    return topk_predictions, topk_labels, label_relevant_counts


def transform_label_to_onehot(labels, vocab_size):
    return tf.one_hot(tf.reshape(labels, (-1,)), vocab_size)


def create_output_placeholder(scores, ks):
    return tf.Variable(tf.zeros([tf.shape(scores)[0], len(ks)], tf.float32))


def gather_torch_like(labels, indices, max_k):

    row_idxs = tf.repeat(tf.range(tf.shape(labels)[0]), max_k)
    col_idx = tf.reshape(indices, tf.shape(row_idxs))
    all_indices = tf.transpose(tf.stack([row_idxs, col_idx]))

    labels = tf.reshape(tf.gather_nd(labels, all_indices), (tf.shape(labels)[0], max_k))
    return labels


def batch_ref(inputs: Union[tf.Tensor, TabularData]):
    """Get hash-code of a tensor or a dictionary of tensors."""

    if isinstance(inputs, tf.Tensor):
        return hash(inputs.ref())

    refs = []
    keys = sorted(inputs.keys())
    for key in keys:
        refs.append(inputs[key].ref())

    return hash(tuple(refs))


def pack_df(gdf):
    if isinstance(gdf, np.ndarray):
        return gdf
    elif hasattr(gdf, "to_dlpack") and callable(getattr(gdf, "to_dlpack")):
        return gdf.to_dlpack()
    elif hasattr(gdf, "to_numpy") and callable(getattr(gdf, "to_numpy")):
        gdf = gdf.to_numpy()
        if isinstance(gdf[0], list):
            gdf = np.stack(gdf)
        return gdf
    return gdf.toDlpack()


def unpack_df(gdf):
    if hasattr(gdf, "shape"):
        return tf.convert_to_tensor(gdf)
    return from_dlpack(gdf)


def df_to_tensor(gdf, dtype=None):
    if gdf.empty:
        return

    # checks necessary because of this bug
    # https://github.com/tensorflow/tensorflow/issues/42660
    if len(gdf.shape) == 1 or gdf.shape[1] == 1:
        dlpack = pack_df(gdf)
    elif gdf.shape[0] == 1:
        dlpack = pack_df(gdf.values[0])
    else:
        dlpack = pack_df(gdf.values.T)
    # catch error caused by tf eager context
    # not being initialized

    try:
        x = unpack_df(dlpack)
    except AssertionError:
        tf.random.uniform((1,))
        x = unpack_df(dlpack)
    # if rank is already two it is  already in list format
    if gdf.shape[0] == 1 and not tf.rank(x) == 2:
        # batch size 1 so got squashed to a vector
        x = tf.expand_dims(x, 0)
    elif len(gdf.shape) == 1 or len(x.shape) == 1:
        # sort of a generic check for any other
        # len(shape)==1 case, could probably
        # be more specific
        x = tf.expand_dims(x, -1)
    elif gdf.shape[1] > 1:
        # matrix which means we had to transpose
        # for the bug above, so untranspose
        x = tf.transpose(x)
    return x


def add_epsilon_to_zeros(tensor: tf.Tensor, epsilon: float = 1e-24) -> tf.Tensor:
    """Replaces zeros by adding a small epsilon value to them.
    This is useful to avoid inf and nan errors on math ops
    like log().

    Parameters
    ----------
    tensor : tf.Tensor
        Tensor to operate on
    epsilon : float, optional
        Small value to add to zeros, by default 1e-24

    Returns
    -------
    tf.Tensor
        The tensor without zeros
    """
    return tf.where(tf.equal(tensor, 0.0), tensor + epsilon, tensor)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import merlin.models.tf as ml
from merlin.io import Dataset


def test_sequential_block_yoochoose(testing_data: Dataset):
    body = ml.InputBlock(testing_data.schema).connect(ml.MLPBlock([64]))

    outputs = body(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert list(outputs.shape) == [100, 64]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Dict, Protocol, Union, runtime_checkable

import tensorflow as tf

from merlin.models.tf.typing import TabularData

if TYPE_CHECKING:
    from merlin.models.tf.blocks.corer.base import PredictionOutput


class LossMixin(abc.ABC):
    """Mixin to use for Keras Layers that can calculate a loss."""

    def compute_loss(
        self,
        inputs: Union[tf.Tensor, TabularData],
        targets: Union[tf.Tensor, TabularData],
        compute_metrics=True,
        training: bool = False,
        **kwargs,
    ) -> tf.Tensor:
        """Compute the loss on a batch of data.

        Parameters
        ----------
        inputs: Union[torch.Tensor, TabularData]
            TODO
        targets: Union[torch.Tensor, TabularData]
            TODO
        training: bool, default=False
        """
        raise NotImplementedError()


class MetricsMixin(abc.ABC):
    """Mixin to use for Keras Layers that can calculate metrics."""

    def calculate_metrics(
        self,
        outputs: PredictionOutput,
        mode: str = "val",
        forward: bool = True,
        training: bool = False,
        **kwargs,
    ) -> Dict[str, Union[Dict[str, tf.Tensor], tf.Tensor]]:

        """Calculate metrics on a batch of data, each metric is stateful and this updates the state.

        The state of each metric can be retrieved by calling the `metric_results` method.

        Parameters
        ----------
        outputs: PredictionOutput
            The named tuple containing predictions and targets tensors
        forward: bool, default True

        mode: str, default="val"

        """
        raise NotImplementedError()

    def metric_results(self, mode: str = "val") -> Dict[str, Union[float, tf.Tensor]]:
        """Returns the current state of each metric.

        The state is typically updated each batch by calling the `calculate_metrics` method.

        Parameters
        ----------
        mode: str, default="val"

        Returns
        -------
        Dict[str, Union[float, tf.Tensor]]
        """
        raise NotImplementedError()

    def reset_metrics(self):
        """Reset all metrics."""
        raise NotImplementedError()


@runtime_checkable
class ModelLikeBlock(Protocol):
    def compute_loss(
        self,
        inputs: Union[tf.Tensor, TabularData],
        targets: Union[tf.Tensor, TabularData],
        compute_metrics=True,
        training: bool = False,
        **kwargs,
    ) -> tf.Tensor:
        ...

    def calculate_metrics(
        self,
        outputs: PredictionOutput,
        mode: str = "val",
        forward=True,
        training=False,
        **kwargs,
    ) -> Dict[str, Union[Dict[str, tf.Tensor], tf.Tensor]]:
        ...

    def metric_results(self, mode: str = "val") -> Dict[str, Union[float, tf.Tensor]]:
        ...

    def _set_context(self, context):
        ...



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tensorflow as tf

import merlin.models.tf as ml
from merlin.io import Dataset

targets = {"target": tf.cast(tf.random.uniform((100,), maxval=2, dtype=tf.int32), tf.float32)}


def test_binary_classification_head(testing_data: Dataset):
    from merlin.models.tf.utils import testing_utils

    body = ml.InputBlock(testing_data.schema).connect(ml.MLPBlock([64]))
    model = body.connect(ml.BinaryClassificationTask("target"))

    batch = (ml.sample_batch(testing_data, batch_size=100, include_targets=False), targets)
    testing_utils.assert_loss_and_metrics_are_valid(model, batch)


def test_serialization_binary_classification_head(testing_data: Dataset):
    from merlin.models.tf.utils import testing_utils

    body = ml.InputBlock(testing_data.schema).connect(ml.MLPBlock([64]))
    model = body.connect(ml.BinaryClassificationTask("target"))

    copy_model = testing_utils.assert_serialization(model)
    batch = (ml.sample_batch(testing_data, batch_size=100, include_targets=False), targets)
    testing_utils.assert_loss_and_metrics_are_valid(copy_model, batch)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from typing import Optional

import tensorflow as tf
from tensorflow.python.layers.base import Layer

from merlin.models.tf.blocks.core.base import Block
from merlin.models.tf.blocks.core.masking import MaskingHead
from merlin.models.tf.blocks.core.transformations import (
    ItemsPredictionWeightTying,
    L2Norm,
    LabelToOneHot,
    LogitsTemperatureScaler,
    RemovePad3D,
)
from merlin.models.tf.blocks.retrieval.base import ItemRetrievalScorer
from merlin.models.tf.blocks.sampling.cross_batch import PopularityBasedSampler
from merlin.models.tf.losses.base import LossType
from merlin.models.tf.metrics.ranking import ranking_metrics
from merlin.models.tf.prediction_tasks.classification import (
    CategFeaturePrediction,
    MultiClassClassificationTask,
)
from merlin.models.utils.schema_utils import categorical_cardinalities
from merlin.schema import Schema, Tags

LOG = logging.getLogger("merlin.models")


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class ItemsPrediction(CategFeaturePrediction):
    def __init__(
        self,
        schema: Schema,
        **kwargs,
    ):
        super(ItemsPrediction, self).__init__(schema, **kwargs)


def ItemsPredictionSampled(
    schema: Schema,
    num_sampled: int,
    min_id: int = 0,
    ignore_false_negatives: bool = True,
):
    """
    Compute the items logits on a subset of sampled candidates to optimize
    training. During inference, the scores are computed over the whole
    catalog of items.
    Reference of the method can be found at [Jean et al., 2014](http://arxiv.org/abs/1412.2007)

    Parameters:
    -----------
        schema: Schema
            The schema object including features to use and their properties.
        num_sampled: int
            The number of candidates to sample during training
        min_id: int
            The minimum id value to be sampled as negative. Useful to ignore the first categorical
            encoded ids, which are usually reserved for <nulls>, out-of-vocabulary or padding.
            Defaults to 0.
        ignore_false_negatives: bool
            Ignore sampled items that are equal to the target classes
            Defaults to True

    Returns:
    -------
        targets, logits: tf.Tensor, tf.Tensor
            During training, return the concatenated tensor of true class
            and sampled negatives of shape (bs, num_sampled+1), as well as the related logits.
            During evaluation, returns the input tensor of true class, and the related logits.
    """
    item_id_feature_name = schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    num_classes = categorical_cardinalities(schema)[item_id_feature_name]
    samplers = PopularityBasedSampler(
        max_num_samples=num_sampled,
        max_id=num_classes,
        min_id=min_id,
        item_id_feature_name=item_id_feature_name,
    )

    logits = ItemRetrievalScorer(
        samplers=[samplers],
        sampling_downscore_false_negatives=ignore_false_negatives,
        item_id_feature_name=item_id_feature_name,
        sampled_softmax_mode=True,
    )

    return logits


def NextItemPredictionTask(
    schema: Schema,
    loss: Optional[LossType] = "categorical_crossentropy",
    metrics=ranking_metrics(top_ks=[10]),
    weight_tying: bool = True,
    masking: bool = True,
    extra_pre_call: Optional[Block] = None,
    target_name: Optional[str] = None,
    task_name: Optional[str] = None,
    task_block: Optional[Layer] = None,
    logits_temperature: float = 1.0,
    normalize: bool = True,
    sampled_softmax: bool = False,
    num_sampled: int = 100,
    min_sampled_id: int = 0,
) -> MultiClassClassificationTask:
    """
    Function to create the NextItemPrediction task with the right parameters.
    Parameters
    ----------
        schema: Schema
            The schema object including features to use and their properties.
        loss: Optional[LossType]
            Loss function.
            Defaults to `sparse_categorical_crossentropy`.
        metrics: MetricOrMetrics
            List of top-k ranking metrics.
            Defaults to ranking_metrics(top_ks=[10]).
        weight_tying: bool
            The item_id embedding weights are shared with the prediction network layer.
            Defaults to True
        masking: bool
            Whether masking is used to transform inputs and targets or not
            Defaults to True
        extra_pre_call: Optional[PredictionBlock]
            Optional extra pre-call block. Defaults to None.
        target_name: Optional[str]
            If specified, name of the target tensor to retrieve from dataloader.
            Defaults to None.
        task_name: Optional[str]
            name of the task.
            Defaults to None.
        task_block: Block
            The `Block` that applies additional layers op to inputs.
            Defaults to None.
        logits_temperature: float
            Parameter used to reduce the model overconfidence, so that logits / T.
            Defaults to 1.
        normalize: bool
            Apply L2 normalization before computing dot interactions.
            Defaults to True.
        sampled_softmax: bool
            Compute the logits scores over all items of the catalog or
            generate a subset of candidates
            Defaults to False
        num_sampled: int
            When sampled_softmax is enabled, specify the number of
            negative candidates to generate for each batch
            Defaults to 100
        min_sampled_id: int
            The minimum id value to be sampled. Useful to ignore the first categorical
            encoded ids, which are usually reserved for <nulls>, out-of-vocabulary or padding.
            Defaults to 0.
    Returns
    -------
        PredictionTask
            The next item prediction task
    """
    item_id_feature_name = schema.select_by_tag(Tags.ITEM_ID).column_names[0]

    if sampled_softmax:
        prediction_call = ItemsPredictionSampled(
            schema, num_sampled=num_sampled, min_id=min_sampled_id
        )

    else:
        if weight_tying:
            prediction_call = ItemsPredictionWeightTying(schema)

        else:
            prediction_call = ItemsPrediction(schema)

        prediction_call = prediction_call.connect(LabelToOneHot())

    if logits_temperature != 1:
        prediction_call = prediction_call.connect(LogitsTemperatureScaler(logits_temperature))

    if masking:
        prediction_call = MaskingHead(item_id_feature_name=item_id_feature_name).connect(
            RemovePad3D(), prediction_call
        )

    if normalize:
        prediction_call = L2Norm().connect(prediction_call)

    if extra_pre_call is not None:
        prediction_call = prediction_call.connect(extra_pre_call)

    return MultiClassClassificationTask(
        target_name,
        task_name,
        task_block,
        loss=loss,
        metrics=metrics,
        pre=prediction_call,
    )



from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from merlin.models.tf.blocks.core.aggregation import SequenceAggregation, SequenceAggregator
from merlin.models.tf.blocks.core.base import Block, BlockType, MetricOrMetrics
from merlin.models.tf.blocks.core.inputs import InputBlock
from merlin.models.tf.blocks.mlp import MLPBlock
from merlin.models.tf.blocks.retrieval.matrix_factorization import QueryItemIdsEmbeddingsBlock
from merlin.models.tf.blocks.retrieval.two_tower import TwoTowerBlock
from merlin.models.tf.blocks.sampling.base import ItemSampler
from merlin.models.tf.features.embedding import EmbeddingOptions
from merlin.models.tf.losses import LossType
from merlin.models.tf.metrics.ranking import ranking_metrics
from merlin.models.tf.models.base import Model, RetrievalModel
from merlin.models.tf.models.utils import parse_prediction_tasks
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.models.tf.prediction_tasks.next_item import NextItemPredictionTask
from merlin.models.tf.prediction_tasks.retrieval import ItemRetrievalTask
from merlin.schema import Schema, Tags


def MatrixFactorizationModel(
    schema: Schema,
    dim: int,
    query_id_tag=Tags.USER_ID,
    item_id_tag=Tags.ITEM_ID,
    embeddings_initializers: Optional[Dict[str, Callable[[Any], None]]] = None,
    post: Optional[BlockType] = None,
    prediction_tasks: Optional[
        Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
    ] = None,
    logits_temperature: float = 1.0,
    loss: Optional[LossType] = "bpr",
    metrics: MetricOrMetrics = ItemRetrievalTask.DEFAULT_METRICS,
    samplers: Sequence[ItemSampler] = (),
    **kwargs,
) -> Union[Model, RetrievalModel]:
    """Builds a matrix factorization model.

    Example Usage::
        mf = MatrixFactorizationModel(schema, dim=128)
        mf.compile(optimizer="adam")
        mf.fit(train_data, epochs=10)

    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    dim: int
        The dimension of the embeddings.
    query_id_tag : Tag
        The tag to select query features, by default `Tags.USER`
    item_id_tag : Tag
        The tag to select item features, by default `Tags.ITEM`
    embeddings_initializers: Dict[str, Callable[[Any], None]]
        A dictionary of initializers for embeddings.
    post: Optional[Block], optional
        The optional `Block` to apply on both outputs of Two-tower model
    prediction_tasks: optional
        The optional `PredictionTask` or list of `PredictionTask` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    loss: Optional[LossType]
        Loss function.
        Defaults to `bpr`.
    samplers: List[ItemSampler]
        List of samplers for negative sampling, by default `[InBatchSampler()]`

    Returns
    -------
    Union[Model, RetrievalModel]
    """

    if not prediction_tasks:
        prediction_tasks = ItemRetrievalTask(
            schema,
            metrics=metrics,
            logits_temperature=logits_temperature,
            samplers=list(samplers),
            loss=loss,
            **kwargs,
        )

    prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
    two_tower = QueryItemIdsEmbeddingsBlock(
        schema=schema,
        dim=dim,
        query_id_tag=query_id_tag,
        item_id_tag=item_id_tag,
        embeddings_initializers=embeddings_initializers,
        post=post,
        **kwargs,
    )

    model = two_tower.connect(prediction_tasks)

    return model


def TwoTowerModel(
    schema: Schema,
    query_tower: Block,
    item_tower: Optional[Block] = None,
    query_tower_tag=Tags.USER,
    item_tower_tag=Tags.ITEM,
    embedding_options: EmbeddingOptions = EmbeddingOptions(
        embedding_dims=None,
        embedding_dim_default=64,
        infer_embedding_sizes=False,
        infer_embedding_sizes_multiplier=2.0,
    ),
    post: Optional[BlockType] = None,
    prediction_tasks: Optional[
        Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
    ] = None,
    logits_temperature: float = 1.0,
    loss: Optional[LossType] = "categorical_crossentropy",
    metrics: MetricOrMetrics = ItemRetrievalTask.DEFAULT_METRICS,
    samplers: Sequence[ItemSampler] = (),
    **kwargs,
) -> Union[Model, RetrievalModel]:
    """Builds the Two-tower architecture, as proposed in [1].

    Example Usage::
        two_tower = TwoTowerModel(schema, MLPBlock([256, 64]))
        two_tower.compile(optimizer="adam")
        two_tower.fit(train_data, epochs=10)

    References
    ----------
    [1] Yi, Xinyang, et al.
        "Sampling-bias-corrected neural modeling for large corpus item recommendations."
        Proceedings of the 13th ACM Conference on Recommender Systems. 2019.

    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    query_tower: Block
        The `Block` that combines user features
    item_tower: Optional[Block], optional
        The optional `Block` that combines items features.
        If not provided, a copy of the query_tower is used.
    query_tower_tag: Tag
        The tag to select query features, by default `Tags.USER`
    item_tower_tag: Tag
        The tag to select item features, by default `Tags.ITEM`
    embedding_options : EmbeddingOptions
        Options for the input embeddings.
        - embedding_dims: Optional[Dict[str, int]] - The dimension of the
        embedding table for each feature (key), by default {}
        - embedding_dim_default: int - Default dimension of the embedding
        table, when the feature is not found in ``embedding_dims``, by default 64
        - infer_embedding_sizes : bool, Automatically defines the embedding
        dimension from the feature cardinality in the schema, by default False
        - infer_embedding_sizes_multiplier: int. Multiplier used by the heuristic
        to infer the embedding dimension from its cardinality. Generally
        reasonable values range between 2.0 and 10.0. By default 2.0.
    post: Optional[Block], optional
        The optional `Block` to apply on both outputs of Two-tower model
    prediction_tasks: optional
        The optional `PredictionTask` or list of `PredictionTask` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    loss: Optional[LossType]
        Loss function.
        Defaults to `categorical_crossentropy`.
    samplers: List[ItemSampler]
        List of samplers for negative sampling, by default `[InBatchSampler()]`

    Returns
    -------
    Union[Model, RetrievalModel]
    """

    if not prediction_tasks:
        prediction_tasks = ItemRetrievalTask(
            schema,
            metrics=metrics,
            logits_temperature=logits_temperature,
            samplers=list(samplers),
            loss=loss,
            # Two-tower outputs are already L2-normalized
            normalize=False,
            **kwargs,
        )

    prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)
    two_tower = TwoTowerBlock(
        schema=schema,
        query_tower=query_tower,
        item_tower=item_tower,
        query_tower_tag=query_tower_tag,
        item_tower_tag=item_tower_tag,
        embedding_options=embedding_options,
        post=post,
        **kwargs,
    )

    model = two_tower.connect(prediction_tasks)

    return model


def YoutubeDNNRetrievalModel(
    schema: Schema,
    max_seq_length: int,
    aggregation: str = "concat",
    top_block: Block = MLPBlock([64]),
    num_sampled: int = 100,
    loss: Optional[LossType] = "categorical_crossentropy",
    metrics=ranking_metrics(top_ks=[10]),
    normalize: bool = True,
    extra_pre_call: Optional[Block] = None,
    task_block: Optional[Block] = None,
    logits_temperature: float = 1.0,
    seq_aggregator: Block = SequenceAggregator(SequenceAggregation.MEAN),
) -> Model:
    """Build the Youtube-DNN retrieval model. More details of the model can be found in [1].

    Example Usage::
        model = YoutubeDNNRetrievalModel(schema, num_sampled=100)
        model.compile(optimizer="adam")
        model.fit(train_data, epochs=10)

    References
    ----------
    [1] Covington, Paul, Jay Adams, and Emre Sargin.
        "Deep neural networks for youtube recommendations."
        Proceedings of the 10th ACM conference on recommender systems. 2016.


    Parameters
    ----------
    schema: Schema
        The `Schema` with the input features
    aggregation: str
        The aggregation method to use for the sequence of features.
        Defaults to `concat`.
    top_block: Block
        The `Block` that combines the top features
    num_sampled: int
        The number of negative samples to use in the sampled-softmax.
        Defaults to 100.
    loss: Optional[LossType]
        Loss function.
        Defaults to `categorical_crossentropy`.
    metrics: List[Metric]
        List of metrics to use.
        Defaults to `ranking_metrics(top_ks=[10])`
    normalize: bool
        Whether to normalize the embeddings.
        Defaults to True.
    extra_pre_call: Optional[Block]
        The optional `Block` to apply before the model.
    task_block: Optional[Block]
        The optional `Block` to apply on the model.
    logits_temperature: float
        Parameter used to reduce model overconfidence, so that logits / T.
        Defaults to 1.
    seq_aggregator: Block
        The `Block` to aggregate the sequence of features.
    """

    inputs = InputBlock(
        schema,
        aggregation=aggregation,
        seq=False,
        max_seq_length=max_seq_length,
        masking="clm",
        split_sparse=True,
        seq_aggregator=seq_aggregator,
    )

    task = NextItemPredictionTask(
        schema=schema,
        loss=loss,
        metrics=metrics,
        masking=True,
        weight_tying=False,
        sampled_softmax=True,
        extra_pre_call=extra_pre_call,
        task_block=task_block,
        logits_temperature=logits_temperature,
        normalize=normalize,
        num_sampled=num_sampled,
    )

    return inputs.connect(top_block, task)



from typing import List

import pytest
import tensorflow as tf
from tensorflow.keras import mixed_precision

import merlin.models.tf as ml
from merlin.io.dataset import Dataset
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


def test_filter_features(tf_con_features):
    features = ["a", "b"]
    con = ml.Filter(features)(tf_con_features)

    assert list(con.keys()) == features


def test_as_tabular(tf_con_features):
    name = "tabular"
    con = ml.AsTabular(name)(tf_con_features)

    assert list(con.keys()) == [name]


def test_tabular_block(tf_con_features):
    _DummyTabular = ml.TabularBlock

    tabular = _DummyTabular()

    assert tabular(tf_con_features) == tf_con_features
    assert tabular(tf_con_features, aggregation="concat").shape[1] == 6
    assert tabular(tf_con_features, aggregation=ml.ConcatFeatures()).shape[1] == 6

    tabular_concat = _DummyTabular(aggregation="concat")
    assert tabular_concat(tf_con_features).shape[1] == 6

    tab_a = ["a"] >> _DummyTabular()
    tab_b = ["b"] >> _DummyTabular()

    assert tab_a(tf_con_features, merge_with=tab_b, aggregation="stack").shape[1] == 1


@pytest.mark.parametrize("pre", [None])
@pytest.mark.parametrize("post", [None])
@pytest.mark.parametrize("aggregation", [None, "concat"])
@pytest.mark.parametrize("include_schema", [True, False])
def test_serialization_continuous_features(
    testing_data: Dataset, pre, post, aggregation, include_schema
):
    schema = None
    if include_schema:
        schema = testing_data.schema

    inputs = ml.TabularBlock(pre=pre, post=post, aggregation=aggregation, schema=schema)

    copy_layer = testing_utils.assert_serialization(inputs)

    keep_cols = ["user_id", "item_id", "event_hour_sin", "event_hour_cos"]
    tf_tabular_data = ml.sample_batch(testing_data, batch_size=100, include_targets=False)
    for k in list(tf_tabular_data.keys()):
        if k not in keep_cols:
            del tf_tabular_data[k]

    assert copy_layer(tf_tabular_data) is not None
    assert inputs.pre.__class__.__name__ == copy_layer.pre.__class__.__name__
    assert inputs.post.__class__.__name__ == copy_layer.post.__class__.__name__
    assert inputs.aggregation.__class__.__name__ == copy_layer.aggregation.__class__.__name__
    if include_schema:
        assert inputs.schema == schema


class DummyFeaturesBlock(ml.Block):
    def add_features_to_context(self, feature_shapes) -> List[str]:
        return [Tags.ITEM_ID.value]

    def call(self, inputs, **kwargs):
        items = self.context[Tags.ITEM_ID]
        emb_table = self.context.get_embedding(Tags.ITEM_ID)
        item_embeddings = tf.gather(emb_table, tf.cast(items, tf.int32))
        if tf.rank(item_embeddings) == 3:
            item_embeddings = tf.squeeze(item_embeddings)

        return inputs * item_embeddings

    def compute_output_shape(self, input_shapes):
        return input_shapes

    @property
    def item_embedding_table(self):
        return self.context.get_embedding(Tags.ITEM_ID)


def test_block_context(ecommerce_data: Dataset):
    inputs = ml.InputBlock(ecommerce_data.schema)
    dummy = DummyFeaturesBlock()
    model = inputs.connect(ml.MLPBlock([64]), dummy, context=ml.BlockContext())
    out = model(ml.sample_batch(ecommerce_data, batch_size=100, include_targets=False))

    embeddings = inputs.select_by_name(Tags.CATEGORICAL.value)
    assert (
        dummy.context.get_embedding(Tags.ITEM_ID).shape
        == embeddings.embedding_tables[Tags.ITEM_ID.value].shape
    )

    assert out.shape[-1] == 64


@pytest.mark.parametrize("run_eagerly", [True])
def test_block_context_model(ecommerce_data: Dataset, run_eagerly: bool, tmp_path):
    dummy = DummyFeaturesBlock()
    model = ml.Model(
        ml.InputBlock(ecommerce_data.schema),
        ml.MLPBlock([64]),
        dummy,
        ml.BinaryClassificationTask("click"),
    )

    model.compile(optimizer="adam", run_eagerly=run_eagerly)
    model.fit(ecommerce_data, batch_size=50, epochs=1)
    model.save(str(tmp_path))

    copy_model = tf.keras.models.load_model(str(tmp_path))
    assert copy_model.context == copy_model.block.layers[0].context
    assert list(copy_model.context._feature_names) == ["item_id"]
    assert len(dict(copy_model.context._feature_dtypes)) == 23

    copy_model.compile(optimizer="adam", run_eagerly=run_eagerly)
    # TODO: Fix prediction-task output name so that we can retrain a model after saving
    # copy_model.fit(ecommerce_data.tf_dataloader(), epochs=1)


def test_simple_model(ecommerce_data: Dataset):
    model = ml.Model(
        ml.InputBlock(ecommerce_data.schema),
        ml.MLPBlock([64]),
        ml.BinaryClassificationTask("click"),
    )

    copy_model = testing_utils.assert_serialization(model)
    testing_utils.assert_loss_and_metrics_are_valid(
        copy_model, ml.sample_batch(ecommerce_data, batch_size=100)
    )


def test_block_to_model(ecommerce_data: Dataset):
    embedding_options = ml.EmbeddingOptions(embedding_dim_default=32)
    model = ml.MLPBlock([64]).to_model(
        ecommerce_data.schema,
        prediction_tasks=ml.BinaryClassificationTask("click"),
        embedding_options=embedding_options,
    )

    copy_model = testing_utils.assert_serialization(model)
    testing_utils.assert_loss_and_metrics_are_valid(
        copy_model, ml.sample_batch(ecommerce_data, batch_size=100)
    )

    assert all(
        [f.table.dim == 32 for f in list(model.block.inputs["categorical"].feature_config.values())]
    )


def test_model_from_block(ecommerce_data: Dataset):
    model = ml.Model.from_block(
        ml.MLPBlock([64]),
        ecommerce_data.schema,
        prediction_tasks=ml.BinaryClassificationTask("click"),
    )

    copy_model = testing_utils.assert_serialization(model)
    testing_utils.assert_loss_and_metrics_are_valid(
        copy_model, ml.sample_batch(ecommerce_data, batch_size=100)
    )


def test_block_with_input_to_model(ecommerce_data: Dataset):
    inputs = ml.InputBlock(ecommerce_data.schema)
    block = inputs.connect(ml.MLPBlock([64]))

    with pytest.raises(ValueError) as excinfo:
        block.to_model(
            ecommerce_data.schema,
            input_block=inputs,
        )
    assert "The block already includes an InputBlock" in str(excinfo.value)


def test_wrong_model(ecommerce_data: Dataset):
    with pytest.raises(ValueError) as excinfo:
        ml.Model(
            ml.InputBlock(ecommerce_data.schema),
            ml.MLPBlock([64]),
        )
    assert "Last block must be able to calculate loss & metrics." in str(excinfo.value)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_block_context_model_fp16(ecommerce_data: Dataset, run_eagerly: bool, num_epochs=2):
    mixed_precision.set_global_policy("mixed_float16")
    model = ml.Model(
        ml.InputBlock(ecommerce_data.schema),
        ml.MLPBlock([32]),
        ml.BinaryClassificationTask("click"),
    )
    model.compile(optimizer="adam", run_eagerly=run_eagerly)
    mixed_precision.set_global_policy("float32")
    losses = model.fit(ecommerce_data, batch_size=100, epochs=2)
    assert len(losses.epoch) == num_epochs
    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])



from mypy.options import Options
from mypy.plugin import FunctionContext, Plugin
from mypy.types import Type


def plugin(version: str):
    return MerlinModelsPlugin


class MerlinModelsPlugin(Plugin):
    def __init__(self, options: Options) -> None:
        super().__init__(options)

    def get_method_hook(self, fullname: str):
        if fullname.startswith("merlin.models.tf") and fullname.endswith("connect"):
            return merlin_connect_hook_tf
        if fullname.startswith("merlin.models.tf") and fullname.endswith("connect_branch"):
            return merlin_connect_hook_tf

        return default_function_hook


def default_function_hook(ctx: FunctionContext) -> Type:
    return ctx.default_return_type


def merlin_connect_hook_tf(ctx: FunctionContext) -> Type:
    SequentialBlock, Model, RetrievalModel = ctx.default_return_type.items  # type: ignore

    model_like_blocks = [
        "merlin.models.tf.core.ParallelPredictionBlock",
        "merlin.models.tf.core.PredictionTask",
        "merlin.models.tf.prediction.classification.BinaryClassificationTask",
        "merlin.models.tf.prediction.regression.RegressionTask",
    ]

    retrieval_blocks = [
        "merlin.models.tf.block.retrieval.TwoTowerBlock",
        "merlin.models.tf.block.retrieval.MatrixFactorizationBlock",
    ]

    if str(ctx.arg_types[0][-1]) in model_like_blocks:
        if str(ctx.type) in retrieval_blocks:  # type: ignore
            return RetrievalModel

        return Model

    return SequentialBlock



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from tensorflow.python.ops import init_ops_v2

import merlin.models.tf as ml
from merlin.io import Dataset
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


def test_embedding_features(tf_cat_features):
    dim = 15
    feature_config = {
        f: ml.FeatureConfig(ml.TableConfig(100, dim, name=f, initializer=None))
        for f in tf_cat_features.keys()
    }
    embeddings = ml.EmbeddingFeatures(feature_config)(tf_cat_features)

    assert list(embeddings.keys()) == list(feature_config.keys())
    assert all([emb.shape[-1] == dim for emb in embeddings.values()])


def test_embedding_features_yoochoose(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = ml.EmbeddingFeatures.from_schema(schema)
    embeddings = emb_module(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert sorted(list(embeddings.keys())) == sorted(schema.column_names)
    assert all(emb.shape[-1] == 64 for emb in embeddings.values())
    max_value = list(schema.select_by_name("item_id"))[0].int_domain.max
    assert emb_module.embedding_tables["item_id"].shape[0] == max_value + 1


def test_serialization_embedding_features(testing_data: Dataset):
    inputs = ml.EmbeddingFeatures.from_schema(testing_data.schema)

    copy_layer = testing_utils.assert_serialization(inputs)

    assert list(inputs.feature_config.keys()) == list(copy_layer.feature_config.keys())

    from merlin.models.tf.features.embedding import serialize_table_config as serialize

    assert all(
        serialize(inputs.feature_config[key].table)
        == serialize(copy_layer.feature_config[key].table)
        for key in copy_layer.feature_config
    )


@testing_utils.mark_run_eagerly_modes
def test_embedding_features_yoochoose_model(music_streaming_data: Dataset, run_eagerly):
    schema = music_streaming_data.schema.select_by_tag(Tags.CATEGORICAL)

    inputs = ml.EmbeddingFeatures.from_schema(schema, aggregation="concat")
    body = ml.SequentialBlock([inputs, ml.MLPBlock([64])])

    testing_utils.assert_body_works_in_model(music_streaming_data, body, run_eagerly)


def test_embedding_features_yoochoose_custom_dims(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = ml.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=ml.EmbeddingOptions(
            embedding_dims={"item_id": 100}, embedding_dim_default=64
        ),
    )

    embeddings = emb_module(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert emb_module.embedding_tables["item_id"].shape[1] == 100
    assert emb_module.embedding_tables["categories"].shape[1] == 64

    assert embeddings["item_id"].shape[1] == 100
    assert embeddings["categories"].shape[1] == 64


def test_embedding_features_yoochoose_infer_embedding_sizes(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = ml.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=ml.EmbeddingOptions(
            infer_embedding_sizes=True, infer_embedding_sizes_multiplier=3.0
        ),
    )

    embeddings = emb_module(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert emb_module.embedding_tables["item_id"].shape[1] == 46
    assert emb_module.embedding_tables["categories"].shape[1] == 13

    assert embeddings["item_id"].shape[1] == 46
    assert embeddings["categories"].shape[1] == 13


def test_embedding_features_yoochoose_custom_initializers(testing_data: Dataset):
    ITEM_MEAN = 1.0
    ITEM_STD = 0.05

    CATEGORY_MEAN = 2.0
    CATEGORY_STD = 0.1

    schema = testing_data.schema.select_by_tag(Tags.CATEGORICAL)
    emb_module = ml.EmbeddingFeatures.from_schema(
        schema,
        embedding_options=ml.EmbeddingOptions(
            embeddings_initializers={
                "item_id": init_ops_v2.TruncatedNormal(mean=ITEM_MEAN, stddev=ITEM_STD),
                "categories": init_ops_v2.TruncatedNormal(mean=CATEGORY_MEAN, stddev=CATEGORY_STD),
            },
        ),
    )

    embeddings = emb_module(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert embeddings["item_id"].numpy().mean() == pytest.approx(ITEM_MEAN, abs=0.1)
    assert embeddings["item_id"].numpy().std() == pytest.approx(ITEM_STD, abs=0.1)

    assert embeddings["categories"].numpy().mean() == pytest.approx(CATEGORY_MEAN, abs=0.1)
    assert embeddings["categories"].numpy().std() == pytest.approx(CATEGORY_STD, abs=0.1)


def test_shared_embeddings(music_streaming_data: Dataset):
    inputs = ml.InputBlock(music_streaming_data.schema)

    embeddings = inputs.select_by_name(Tags.CATEGORICAL.value)

    assert embeddings.table_config("item_genres") == embeddings.table_config("user_genres")



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from merlin.schema import Schema, Tags


class SchemaMixin:
    REQUIRES_SCHEMA = False

    def set_schema(self, schema=None):
        self.check_schema(schema=schema)

        if schema and not getattr(self, "_schema", None):
            self._schema = schema

        return self

    @property
    def schema(self) -> Schema:
        if not self.has_schema:
            raise ValueError(f"{self.__class__.__name__} does not have a schema.")

        return self._schema

    @property
    def has_schema(self):
        return getattr(self, "_schema", None) is not None

    @schema.setter  # type: ignore
    def schema(self, value):
        if value:
            self.set_schema(value)
        else:
            self._schema = value

    def check_schema(self, schema=None):
        if self.REQUIRES_SCHEMA and not getattr(self, "_schema", None) and not schema:
            raise ValueError(f"{self.__class__.__name__} requires a schema.")

    def __call__(self, *args, **kwargs):
        self.check_schema()

        return super().__call__(*args, **kwargs)

    def _maybe_set_schema(self, input, schema):
        if input and getattr(input, "set_schema", None):
            input.set_schema(schema)

    def get_item_ids_from_inputs(self, inputs):
        return inputs[self.schema.select_by_tag(Tags.ITEM_ID).first.name]

    def get_padding_mask_from_item_id(self, inputs, pad_token=0):
        item_id_inputs = self.get_item_ids_from_inputs(inputs)
        if len(item_id_inputs.shape) != 2:
            raise ValueError(
                "To extract the padding mask from item id tensor "
                "it is expected to have 2 dims, but it has {} dims.".format(item_id_inputs.shape)
            )
        return self.get_item_ids_from_inputs(inputs) != pad_token


def requires_schema(module):
    module.REQUIRES_SCHEMA = True

    return module



import inspect
from collections import Sequence as SequenceCollection
from collections import defaultdict
from typing import Dict, List, Optional, Sequence, Text, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.python.keras.utils import generic_utils

from merlin.models.tf.blocks.core.base import (
    Block,
    BlockContext,
    BlockType,
    ContextMixin,
    MetricOrMetricClass,
    MetricOrMetrics,
    PredictionOutput,
    _output_metrics,
    name_fn,
)
from merlin.models.tf.blocks.core.combinators import ParallelBlock
from merlin.models.tf.metrics.ranking import RankingMetric
from merlin.models.tf.typing import TabularData, TensorOrTabularData
from merlin.models.tf.utils import tf_utils
from merlin.models.tf.utils.mixins import LossMixin, MetricsMixin
from merlin.models.utils.misc_utils import filter_kwargs
from merlin.models.utils.schema_utils import tensorflow_metadata_json_to_schema
from merlin.schema import Schema, Tags


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class PredictionTask(Layer, LossMixin, MetricsMixin, ContextMixin):
    """Base-class for prediction tasks.

    Parameters
    ----------
    target_name : Optional[str], optional
        Label name, by default None
    task_name : Optional[str], optional
        Task name, by default None
    metrics : Optional[MetricOrMetrics], optional
        List of Keras metrics to be evaluated, by default None
    pre : Optional[Block], optional
        Optional block to transform predictions before computing loss and metrics,
        by default None
    pre_eval_topk : Optional[Block], optional
        Optional block to apply additional transform predictions before computing
        top-k evaluation loss and metrics, by default None
    task_block : Optional[Layer], optional
        Optional block to apply to inputs before computing predictions,
        by default None
    prediction_metrics : Optional[List[tf.keras.metrics.Metric]], optional
        List of Keras metrics used to summarize the predictions, by default None
    label_metrics : Optional[List[tf.keras.metrics.Metric]], optional
        List of Keras metrics used to summarize the labels, by default None
    loss_metrics : Optional[List[tf.keras.metrics.Metric]], optional
        List of Keras metrics used to summarize the loss, by default None
    compute_train_metrics : Optional[bool], optional
        Enable computing metrics during training, by default True
    name : Optional[Text], optional
        Task name, by default None
    """

    def __init__(
        self,
        target_name: Optional[str] = None,
        task_name: Optional[str] = None,
        metrics: Optional[MetricOrMetrics] = None,
        pre: Optional[Block] = None,
        pre_eval_topk: Optional[Block] = None,
        task_block: Optional[Layer] = None,
        prediction_metrics: Optional[List[tf.keras.metrics.Metric]] = None,
        label_metrics: Optional[List[tf.keras.metrics.Metric]] = None,
        loss_metrics: Optional[List[tf.keras.metrics.Metric]] = None,
        compute_train_metrics: Optional[bool] = True,
        name: Optional[Text] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.target_name = target_name
        self.task_block = task_block
        self._task_name = task_name
        self.pre = pre
        self._pre_eval_topk = pre_eval_topk

        if metrics is not None:
            if isinstance(metrics, SequenceCollection):
                metrics = list(metrics)
            else:
                metrics = [metrics]

        create_metrics = self._create_metrics
        self.eval_metrics = create_metrics(metrics) if metrics else []
        self.prediction_metrics = create_metrics(prediction_metrics) if prediction_metrics else []
        self.label_metrics = create_metrics(label_metrics) if label_metrics else []
        self.loss_metrics = create_metrics(loss_metrics) if loss_metrics else []
        self.compute_train_metrics = compute_train_metrics

    @property
    def pre_eval_topk(self):
        return self._pre_eval_topk

    @pre_eval_topk.setter
    def pre_eval_topk(self, value: Block):
        """Set pre_eval_topk Block

        Parameters
        ----------
        value : Block
            The block for top-k evaluation
        """
        self._pre_eval_topk = value

    def pre_call(self, inputs: TensorOrTabularData, **kwargs) -> tf.Tensor:
        """Apply PredictionTask to inputs to get predictions scores

        Parameters
        ----------
        inputs : TensorOrTabularData
            inputs of the prediction task

        Returns
        -------
        tf.Tensor
        """
        x = inputs

        if self.task_block:
            x = self.task_block(x)

        if self.pre:
            x = self.pre(inputs, **kwargs)

        return x

    def pre_loss(self, outputs: PredictionOutput, **kwargs) -> "PredictionOutput":
        """Apply `call_outputs` method of `pre` block to transform predictions and targets
        before computing loss and metrics.

        Parameters
        ----------
        outputs : PredictionOutput
            The named tuple containing predictions and targets tensors

        Returns
        -------
        PredictionOutput
             The named tuple containing transformed predictions and targets tensors
        """
        out = self.pre.call_outputs(outputs, **kwargs) if self.pre else outputs

        return out

    def __call__(self, *args, **kwargs):
        inputs = self.pre_call(*args, **kwargs)

        # This will call the `call` method implemented by the super class.
        outputs = super().__call__(inputs, **kwargs)  # noqa

        return outputs

    def build_task(self, input_shape, schema: Schema, body: Block, **kwargs):
        return super().build(input_shape)

    def _create_metrics(
        self, metrics: Sequence[MetricOrMetricClass]
    ) -> List[tf.keras.metrics.Metric]:
        outputs = []
        for metric in metrics:
            if not isinstance(metric, tf.keras.metrics.Metric):
                metric = metric(name=self.child_name(generic_utils.to_snake_case(metric.__name__)))
            outputs.append(metric)

        return outputs

    @property
    def task_name(self):
        if self._task_name:
            return self._task_name

        base_name = generic_utils.to_snake_case(self.__class__.__name__)

        return name_fn(self.target_name, base_name) if self.target_name else base_name

    def child_name(self, name):
        return name_fn(self.task_name, name)

    def _compute_loss(
        self,
        outputs: PredictionOutput,
        sample_weight: tf.Tensor = None,
        training: bool = False,
        **kwargs,
    ) -> tf.Tensor:
        if "sample_weight" in inspect.signature(self.loss).parameters:
            kwargs["sample_weight"] = sample_weight
        if "valid_negatives_mask" in inspect.signature(self.loss).parameters:
            kwargs["valid_negatives_mask"] = outputs.valid_negatives_mask

        return self.loss(outputs.targets, outputs.predictions, **kwargs)

    def compute_loss(  # type: ignore
        self,
        predictions: tf.Tensor,
        targets: tf.Tensor,
        training: bool = False,
        compute_metrics=False,
        sample_weight: Optional[tf.Tensor] = None,
        **kwargs,
    ) -> tf.Tensor:
        """Method to compute the loss and metrics during
        training/evaluation.


        Parameters
        ----------
        predictions : tf.Tesor
            The tensor of prediction scores.
        targets : _type_
            The tensor of true labels.
        training : bool, optional
            Compute loss in `training` mode,
            by default False.
        compute_metrics : bool, optional
            Compute metrics in addition to the loss,
            by default False

        Returns
        -------
        tf.Tensor
        """
        if isinstance(targets, dict) and self.target_name:
            targets = targets[self.target_name]

        if isinstance(predictions, dict) and self.target_name and self.task_name in predictions:
            predictions = predictions[self.task_name]

        prediction_output = PredictionOutput(predictions, targets)

        if self.pre:
            prediction_output = self.pre_loss(prediction_output, training=training, **kwargs)

        if (
            isinstance(prediction_output.targets, tf.Tensor)
            and len(prediction_output.targets.shape) == len(prediction_output.predictions.shape) - 1
        ):
            prediction_output = prediction_output.copy_with_updates(
                predictions=tf.squeeze(prediction_output.predictions)
            )

        if not training and self._pre_eval_topk:
            # During eval, the retrieval-task only returns positive scores
            # so we need to retrieve top-k negative scores to compute the loss
            prediction_output = self._pre_eval_topk.call_outputs(prediction_output, **kwargs)

        loss = self._compute_loss(
            prediction_output,
            sample_weight=sample_weight,
            training=training,
        )

        loss = tf.cond(
            tf.convert_to_tensor(compute_metrics),
            lambda: self.attach_metrics_calculation_to_loss(prediction_output, loss, training),
            lambda: loss,
        )

        return loss

    def attach_metrics_calculation_to_loss(
        self, outputs: PredictionOutput, loss: tf.Tensor, training: bool
    ):
        update_ops = self.calculate_metrics(outputs, loss=loss, forward=False, training=training)

        update_ops = [x for x in update_ops if x is not None]

        with tf.control_dependencies(update_ops):
            return tf.identity(loss)

    def repr_add(self):
        return [("loss", self.loss)]

    def calculate_metrics(
        self,
        outputs: PredictionOutput,
        sample_weight: Optional[tf.Tensor] = None,
        forward: Optional[bool] = True,
        loss: Optional[tf.Tensor] = None,
        training: Optional[bool] = False,
        **kwargs,
    ):
        """Method to compute the metrics.

        Parameters
        ----------
        outputs : PredictionOutput
            The named tuple containing predictions and targets tensors
        forward : bool, optional
            Apply the `call` of Prediction Task, by default True
        loss : tf.Tensor, optional
            The loss value to attach the metric to, by default None
        training : bool, optional
            Compute metrics in `training` mode,
            by default False

        Returns
        -------
        _type_
            _description_

        Raises
        ------
        Exception
            _description_
        """
        predictions, targets, label_relevant_counts_eval = (
            outputs.predictions,
            outputs.targets,
            outputs.label_relevant_counts,
        )
        if isinstance(targets, dict) and self.target_name:
            targets = targets[self.target_name]

        if forward:
            predictions = self(predictions)

        update_ops = []
        if len(self.eval_metrics) > 0:
            predictions_eval = predictions
            targets_eval = targets

            ranking_metrics = list(
                [metric for metric in self.eval_metrics if isinstance(metric, RankingMetric)]
            )

            if len(ranking_metrics) > 0:

                tf.assert_equal(
                    tf.shape(targets),
                    tf.shape(predictions),
                    f"Predictions ({tf.shape(predictions)}) and targets ({tf.shape(targets)}) "
                    f"should have the same shape. Check if targets were one-hot encoded "
                    f"(with LabelToOneHot() block for example).",
                )

                max_k = tf.reduce_max([metric.k for metric in ranking_metrics])
                tf.debugging.assert_greater_equal(
                    tf.shape(predictions_eval)[-1],
                    max_k,
                    f"The max k for ranking metrics ({max_k}) is higher than "
                    f"the number of predictions in this batch",
                )

                pre_sorted_ranking_metrics = list(
                    [metric for metric in ranking_metrics if metric.pre_sorted]
                )

                if len(pre_sorted_ranking_metrics) > 0:

                    if len(ranking_metrics) > len(pre_sorted_ranking_metrics):
                        raise Exception(
                            "If one of the ranking metrics is set to 'pre_sorted=True', all the "
                            "other ranking metrics should have the same configuration to avoid "
                            "sorting predictions twice."
                        )

                    if training or self._pre_eval_topk is None:
                        # Checking the highest k used in ranking metrics
                        max_k = tf.reduce_max([metric.k for metric in pre_sorted_ranking_metrics])
                        # Pre-sorts predictions and targets (by prediction scores) only once
                        # for all ranking metric (for performance optimization) and extracts
                        # only the top-k predictions/targets.
                        # The label_relevant_counts is necessary because when extracting the labels
                        # from the top-k predictions some relevant items might not be included, but
                        # the relevant count is necessary for many ranking metrics (recall, ndcg)
                        (
                            predictions_eval,
                            targets_eval,
                            label_relevant_counts_eval,
                        ) = tf_utils.extract_topk(max_k, predictions, targets)

            for metric in self.eval_metrics:
                if isinstance(metric, RankingMetric) and metric.pre_sorted:
                    metric_state = metric.update_state(
                        targets_eval,
                        predictions_eval,
                        label_relevant_counts_eval,
                        sample_weight=sample_weight,
                    )
                else:
                    metric_state = metric.update_state(
                        y_true=targets, y_pred=predictions, sample_weight=sample_weight
                    )
                update_ops.append(metric_state)

        for metric in self.prediction_metrics:
            update_ops.append(metric.update_state(predictions, sample_weight=sample_weight))

        for metric in self.label_metrics:
            update_ops.append(metric.update_state(targets, sample_weight=sample_weight))

        for metric in self.loss_metrics:
            if not loss:
                loss = self.loss(y_true=targets, y_pred=predictions, sample_weight=sample_weight)
            update_ops.append(metric.update_state(loss, sample_weight=sample_weight))

        return update_ops

    def metric_results(self, mode: str = "val"):
        """Computes and returns the  computed metric values

        Parameters
        ----------
        mode : str, optional
            string prefix to indicate the mode used to compute metrics,
            by default "val"

        Returns
        -------
        dict
            dictionary of scalar metrics
        """
        return {metric.name: metric.result() for metric in self.metrics}

    def metric_result_dict(self, mode=None):
        return self.metric_results(mode=mode)

    def reset_metrics(self):
        for metric in self.metrics:
            metric.reset()

    @classmethod
    def from_config(cls, config):
        config = tf_utils.maybe_deserialize_keras_objects(
            config,
            {
                "pre": tf.keras.layers.deserialize,
                "metrics": tf.keras.metrics.deserialize,
                "prediction_metrics": tf.keras.metrics.deserialize,
                "label_metrics": tf.keras.metrics.deserialize,
                "loss_metrics": tf.keras.metrics.deserialize,
            },
        )

        return super().from_config(config)

    def get_config(self):
        config = super().get_config()
        config = tf_utils.maybe_serialize_keras_objects(
            self,
            config,
            ["metrics", "prediction_metrics", "label_metrics", "loss_metrics", "pre"],
        )

        # config["summary_type"] = self.sequence_summary.summary_type
        if self.target_name:
            config["target_name"] = self.target_name
        if self._task_name:
            config["task_name"] = self._task_name

        if "metrics" not in config:
            config["metrics"] = []

        return config


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ParallelPredictionBlock(ParallelBlock, LossMixin, MetricsMixin):
    """Multi-task prediction block.

    Parameters
    ----------
    prediction_tasks: *PredictionTask
        List of tasks to be used for prediction.
    task_blocks: Optional[Union[Layer, Dict[str, Layer]]]
        Task blocks to be used for prediction.
    task_weights : Optional[List[float]]
        Weights for each task.
    bias_block : Optional[Layer]
        Bias block to be used for prediction.
    loss_reduction : Callable
        Reduction function for loss.

    """

    def __init__(
        self,
        *prediction_tasks: PredictionTask,
        task_blocks: Optional[Union[Layer, Dict[str, Layer]]] = None,
        task_weights: Optional[List[float]] = None,
        bias_block: Optional[Layer] = None,
        loss_reduction=tf.reduce_mean,
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        **kwargs,
    ):
        self.loss_reduction = loss_reduction

        self.prediction_tasks = prediction_tasks
        self.task_weights = task_weights

        self.bias_block = bias_block
        self.bias_logit = tf.keras.layers.Dense(1)

        self.prediction_task_dict = {}
        if prediction_tasks:
            for task in prediction_tasks:
                self.prediction_task_dict[task.task_name] = task

        super(ParallelPredictionBlock, self).__init__(self.prediction_task_dict, pre=pre, post=post)

        self._task_weight_dict = defaultdict(lambda: 1.0)
        if task_weights:
            for task, val in zip(prediction_tasks, task_weights):
                self._task_weight_dict[task.task_name] = val

        self._set_task_blocks(task_blocks)

    @classmethod
    def get_tasks_from_schema(cls, schema, task_weight_dict=None):
        task_weight_dict = task_weight_dict or {}

        tasks: List[PredictionTask] = []
        task_weights = []
        from merlin.models.tf.prediction_tasks.classification import BinaryClassificationTask
        from merlin.models.tf.prediction_tasks.regression import RegressionTask

        for binary_target in schema.select_by_tag(Tags.BINARY_CLASSIFICATION).column_names:
            tasks.append(BinaryClassificationTask(binary_target))
            task_weights.append(task_weight_dict.get(binary_target, 1.0))
        for regression_target in schema.select_by_tag(Tags.REGRESSION).column_names:
            tasks.append(RegressionTask(regression_target))
            task_weights.append(task_weight_dict.get(regression_target, 1.0))
        # TODO: Add multi-class classification here. Figure out how to get number of classes

        return task_weights, tasks

    @classmethod
    def from_schema(  # type: ignore
        cls,
        schema: Schema,
        task_blocks: Optional[Union[Layer, Dict[str, Layer]]] = None,
        task_weight_dict: Optional[Dict[str, float]] = None,
        bias_block: Optional[Layer] = None,
        loss_reduction=tf.reduce_mean,
        **kwargs,
    ) -> "ParallelPredictionBlock":
        """Built Multi-task prediction Block from schema

        Parameters
        ----------
        schema : Schema
            The `Schema` with the input features
        task_blocks : Optional[Union[Layer, Dict[str, Layer]]], optional
            Task blocks to be used for prediction, by default None
        task_weight_dict : Optional[Dict[str, float]], optional
            Weights for each task, by default None
        bias_block : Optional[Layer], optional
            Bias block to be used for prediction, by default None
        loss_reduction : _type_, optional
            Reduction function for loss, by default tf.reduce_mean
        """
        task_weight_dict = task_weight_dict or {}

        task_weights, tasks = cls.get_tasks_from_schema(schema, task_weight_dict)

        return cls(
            *tasks,
            task_blocks=task_blocks,
            task_weights=task_weights,
            bias_block=bias_block,
            loss_reduction=loss_reduction,
            **kwargs,
        )

    @classmethod
    def task_names_from_schema(cls, schema: Schema) -> List[str]:
        _, tasks = cls.get_tasks_from_schema(schema)

        return [task.task_name for task in tasks]

    def _set_task_blocks(self, task_blocks):
        if not task_blocks:
            return

        if isinstance(task_blocks, dict):
            tasks_multi_names = self._prediction_tasks_multi_names()
            for key, task_block in task_blocks.items():
                if key in tasks_multi_names:
                    tasks = tasks_multi_names[key]
                    if len(tasks) == 1:
                        self.prediction_task_dict[tasks[0].task_name].task_block = task_block
                    else:
                        raise ValueError(
                            f"Ambiguous name: {key}, can't resolve it to a task "
                            "because there are multiple tasks that contain the key: "
                            f"{', '.join([task.task_name for task in tasks])}"
                        )
                else:
                    raise ValueError(
                        f"Couldn't find {key} in prediction_tasks, "
                        f"only found: {', '.join(list(self.prediction_task_dict.keys()))}"
                    )
        elif isinstance(task_blocks, Layer):
            for key, val in self.prediction_task_dict.items():
                task_block = task_blocks.from_config(task_blocks.get_config())
                val.task_block = task_block
        else:
            raise ValueError("`task_blocks` must be a Layer or a Dict[str, Layer]")

    def _prediction_tasks_multi_names(self) -> Dict[str, List[PredictionTask]]:
        prediction_tasks_multi_names = {
            name: [val] for name, val in self.prediction_task_dict.items()
        }
        for name, value in self.prediction_task_dict.items():
            name_parts = name.split("/")
            for name_part in name_parts:
                if name_part in prediction_tasks_multi_names:
                    prediction_tasks_multi_names[name_part].append(value)
                else:
                    prediction_tasks_multi_names[name_part] = [value]

        return prediction_tasks_multi_names

    def add_task(self, task: PredictionTask, task_weight=1):
        key = task.target_name
        if not key:
            raise ValueError("PredictionTask must have a target_name")
        self.parallel_dict[key] = task
        if task_weight:
            self._task_weight_dict[key] = task_weight

        return self

    def pop_labels(self, inputs: Dict[Text, tf.Tensor]):
        outputs = {}
        for name in self.parallel_dict.keys():
            outputs[name] = inputs.pop(name)

        return outputs

    def call(
        self,
        inputs: Union[TabularData, tf.Tensor],
        training: bool = False,
        bias_outputs=None,
        **kwargs,
    ):
        if isinstance(inputs, dict) and not all(
            name in inputs for name in list(self.parallel_dict.keys())
        ):
            if self.bias_block and not bias_outputs:
                bias_outputs = self.bias_block(inputs)
            inputs = self.body(inputs)

        outputs = super(ParallelPredictionBlock, self).call(inputs, **kwargs)

        if bias_outputs is not None:
            for key in outputs:
                outputs[key] += bias_outputs

        return outputs

    def compute_call_output_shape(self, input_shape):
        if isinstance(input_shape, dict) and not all(
            name in input_shape for name in list(self.parallel_dict.keys())
        ):
            input_shape = self.body.compute_output_shape(input_shape)

        return super().compute_call_output_shape(input_shape)

    def compute_loss(  # type: ignore
        self, inputs: Union[tf.Tensor, TabularData], targets, training=False, **kwargs
    ) -> tf.Tensor:
        losses = []

        if isinstance(inputs, dict) and not all(
            name in inputs for name in list(self.parallel_dict.keys())
        ):
            filtered_kwargs = filter_kwargs(
                dict(training=training), self, filter_positional_or_keyword=False
            )
            predictions = self(inputs, **filtered_kwargs)
        else:
            predictions = inputs

        for name, task in self.prediction_task_dict.items():
            loss = task.compute_loss(predictions, targets, training=training, **kwargs)
            losses.append(loss * self._task_weight_dict[name])

        return self.loss_reduction(losses)

    def metric_results(self, mode=None):
        def name_fn(x):
            return "_".join([mode, x]) if mode else x

        metrics = {
            name_fn(name): task.metric_results() for name, task in self.prediction_task_dict.items()
        }

        return _output_metrics(metrics)

    def metric_result_dict(self, mode=None):
        results = {}
        for name, task in self.prediction_task_dict.items():
            results.update(task.metric_results(mode=mode))

        return results

    def reset_metrics(self):
        for task in self.prediction_task_dict.values():
            task.reset_metrics()

    @property
    def task_blocks(self) -> Dict[str, Optional[Layer]]:
        return {name: task.task_block for name, task in self.prediction_task_dict.items()}

    @property
    def task_names(self) -> List[str]:
        return [name for name in self.prediction_task_dict]

    @property
    def metrics(self) -> Dict[str, tf.keras.metrics.Metric]:
        outputs = {}
        for name, task in self.parallel_dict.items():
            outputs.update({metric.name: metric for metric in task.metrics})

        return outputs

    def repr_ignore(self) -> List[str]:
        return ["prediction_tasks", "parallel_layers"]

    def _set_context(self, context: "BlockContext"):
        for task in self.prediction_task_dict.values():
            task._set_context(context)
        super(ParallelPredictionBlock, self)._set_context(context)

    @classmethod
    def from_config(cls, config, **kwargs):
        config = tf_utils.maybe_deserialize_keras_objects(config, ["body", "prediction_tasks"])

        if "schema" in config:
            config["schema"] = tensorflow_metadata_json_to_schema(config["schema"])

        config["loss_reduction"] = getattr(tf, config["loss_reduction"])

        prediction_tasks = config.pop("prediction_tasks", [])

        return cls(*prediction_tasks, **config)

    def get_config(self):
        config = super().get_config()
        config = tf_utils.maybe_serialize_keras_objects(
            self, config, ["body", "loss_reduction", "prediction_tasks"]
        )
        if self.task_weights:
            config["task_weights"] = self.task_weights

        return config



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import platform
import tempfile

import pytest
import tensorflow as tf

from merlin.models.tf.models.base import Model
from merlin.models.tf.prediction_tasks.classification import BinaryClassificationTask


def mark_run_eagerly_modes(*args, **kwargs):
    modes = [True, False]

    # As of TF 2.5 there's a bug that our EmbeddingFeatures don't work on M1 Macs
    if "macOS" in platform.platform() and "arm64-arm-64bit" in platform.platform():
        modes = [True]

    return pytest.mark.parametrize("run_eagerly", modes)(*args, **kwargs)


def assert_body_works_in_model(dataset, body, run_eagerly, num_epochs=5):
    model = body.connect(BinaryClassificationTask("click"))
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    losses = model.fit(dataset, batch_size=50, epochs=num_epochs)
    metrics = model.evaluate(dataset, batch_size=50, return_dict=True)

    assert_binary_classification_loss_metrics(
        losses, metrics, target_name="click", num_epochs=num_epochs
    )


def assert_binary_classification_loss_metrics(losses, metrics, target_name, num_epochs):
    metrics_names = [
        f"{target_name}/binary_classification_task/precision",
        f"{target_name}/binary_classification_task/recall",
        f"{target_name}/binary_classification_task/binary_accuracy",
        f"{target_name}/binary_classification_task/auc",
        "loss",
        "regularization_loss",
        "total_loss",
    ]

    assert len(set(metrics.keys()).intersection(set(metrics_names))) == len(metrics_names)

    assert len(set(losses.history.keys()).intersection(set(metrics_names))) == len(metrics_names)
    assert len(losses.epoch) == num_epochs
    for metric in losses.history.keys():
        assert type(losses.history[metric]) is list
        assert len(losses.history[metric]) == num_epochs

    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])


def assert_regression_loss_metrics(losses, metrics, target_name, num_epochs):
    metrics_names = [
        f"{target_name}/regression_task/root_mean_squared_error",
        "loss",
        "regularization_loss",
        "total_loss",
    ]

    assert len(set(metrics.keys()).intersection(set(metrics_names))) == len(metrics_names)

    assert len(set(losses.history.keys()).intersection(set(metrics_names))) == len(metrics_names)
    assert len(losses.epoch) == num_epochs
    for metric in losses.history.keys():
        assert type(losses.history[metric]) is list
        assert len(losses.history[metric]) == num_epochs

    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])


def assert_loss_and_metrics_are_valid(
    loss_block, features_and_targets, call_body=True, training=False
):
    features, targets = features_and_targets
    predictions = loss_block(features, training=training)
    loss = loss_block.compute_loss(predictions, targets, call_body=call_body, training=training)
    # metrics = input.metric_results()

    assert loss is not None
    # assert len(metrics) == len(input.metrics)


def assert_serialization(layer):
    serialized = layer.get_config()
    copy_layer = layer.from_config(serialized)

    assert isinstance(copy_layer, layer.__class__)

    return copy_layer


def assert_model_is_retrainable(
    model: Model, data, run_eagerly: bool = True, optimizer="adam", **kwargs
):
    model.compile(run_eagerly=run_eagerly, optimizer=optimizer, **kwargs)
    losses = model.fit(data, batch_size=50, epochs=1)

    assert len(losses.epoch) == 1
    # assert all(0 <= loss <= 1 for loss in losses.history["loss"])

    assert model.from_config(model.get_config()) is not None

    with tempfile.TemporaryDirectory() as tmpdir:
        model.save(tmpdir)
        loaded_model = tf.keras.models.load_model(tmpdir)

    assert isinstance(loaded_model, Model)
    loaded_model.compile(run_eagerly=run_eagerly, optimizer=optimizer, **kwargs)
    losses = loaded_model.fit(data, batch_size=50, epochs=1)

    assert len(losses.epoch) == 1
    # assert all(0 <= loss <= 1 for loss in losses.history["loss"])

    return loaded_model



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

import merlin.models.tf as ml
from merlin.io.dataset import Dataset
from merlin.schema import Tags


def test_topk_index(ecommerce_data: Dataset):
    model: ml.RetrievalModel = ml.TwoTowerModel(
        ecommerce_data.schema, query_tower=ml.MLPBlock([64, 128])
    )
    model.compile(run_eagerly=True, optimizer="adam")
    model.fit(ecommerce_data, epochs=1, batch_size=50)

    item_features = ecommerce_data.schema.select_by_tag(Tags.ITEM).column_names
    item_dataset = ecommerce_data.to_ddf()[item_features].drop_duplicates().compute()
    item_dataset = Dataset(item_dataset)

    recommender = model.to_top_k_recommender(item_dataset, k=20)

    batch = ml.sample_batch(ecommerce_data, batch_size=10, include_targets=False)
    _, top_indices = recommender(batch)
    assert top_indices.shape[-1] == 20
    _, top_indices = recommender(batch, k=10)
    assert top_indices.shape[-1] == 10


def test_topk_index_duplicate_indices(ecommerce_data: Dataset):
    model: ml.RetrievalModel = ml.TwoTowerModel(
        ecommerce_data.schema, query_tower=ml.MLPBlock([64, 128])
    )
    model.compile(run_eagerly=True, optimizer="adam")
    model.fit(ecommerce_data, epochs=1, batch_size=50)
    item_features = ecommerce_data.schema.select_by_tag(Tags.ITEM).column_names
    item_dataset = ecommerce_data.to_ddf()[item_features].compute()
    item_dataset = Dataset(item_dataset)

    with pytest.raises(ValueError) as excinfo:
        _ = model.to_top_k_recommender(item_dataset, k=20)
    assert "Please make sure that `data` contains unique indices" in str(excinfo.value)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from typing import List, Optional, Sequence, Union

import tensorflow as tf
from tensorflow.python.ops import embedding_ops

from merlin.models.tf.blocks.core.base import (
    Block,
    BlockType,
    EmbeddingWithMetadata,
    PredictionOutput,
)
from merlin.models.tf.blocks.core.combinators import ParallelBlock
from merlin.models.tf.blocks.core.tabular import Filter, TabularAggregationType
from merlin.models.tf.blocks.sampling.base import ItemSampler
from merlin.models.tf.models.base import ModelBlock
from merlin.models.tf.typing import TabularData
from merlin.models.tf.utils.tf_utils import (
    maybe_deserialize_keras_objects,
    maybe_serialize_keras_objects,
    rescore_false_negatives,
)
from merlin.models.utils.constants import MIN_FLOAT
from merlin.schema import Schema

LOG = logging.getLogger("merlin_models")


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class TowerBlock(ModelBlock):
    """TowerBlock to wrap item or query tower"""

    pass


class RetrievalMixin:
    def query_block(self) -> TowerBlock:
        """Method to return the query tower from a RetrievalModel instance"""
        raise NotImplementedError()

    def item_block(self) -> TowerBlock:
        """Method to return the item tower from a RetrievalModel instance"""
        raise NotImplementedError()


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class DualEncoderBlock(ParallelBlock):
    def __init__(
        self,
        query_block: Block,
        item_block: Block,
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        schema: Optional[Schema] = None,
        name: Optional[str] = None,
        strict: bool = False,
        **kwargs,
    ):
        """Prepare the Query and Item towers of a Retrieval block

        Parameters
        ----------
        query_block : Block
            The `Block` instance that combines user features
        item_block : Block
            Optional `Block` instance that combines items features.
        pre : Optional[BlockType], optional
            Optional `Block` instance to apply before the `call` method of the Two-Tower block
        post : Optional[BlockType], optional
            Optional `Block` instance to apply on both outputs of Two-tower model
        aggregation : Optional[TabularAggregationType], optional
            The Aggregation operation to apply after processing the `call` method
            to output a single Tensor.
        schema : Optional[Schema], optional
            The `Schema` object with the input features.
        name : Optional[str], optional
            Name of the layer.
        strict : bool, optional
            If enabled, check that the input of the ParallelBlock instance is a dictionary.
        """
        self._query_block = TowerBlock(query_block)
        self._item_block = TowerBlock(item_block)

        branches = {
            "query": Filter(query_block.schema).connect(self._query_block),
            "item": Filter(item_block.schema).connect(self._item_block),
        }

        super().__init__(
            branches,
            pre=pre,
            post=post,
            aggregation=aggregation,
            schema=schema,
            name=name,
            strict=strict,
            **kwargs,
        )

    def query_block(self) -> TowerBlock:
        return self._query_block

    def item_block(self) -> TowerBlock:
        return self._item_block

    @classmethod
    def from_config(cls, config, custom_objects=None):
        inputs, config = cls.parse_config(config, custom_objects)
        output = ParallelBlock(inputs, **config)
        output.__class__ = cls

        return output


@Block.registry.register_with_multiple_names("item_retrieval_scorer")
@tf.keras.utils.register_keras_serializable(package="merlin_models")
class ItemRetrievalScorer(Block):
    """Block for ItemRetrieval, which expects query/user and item embeddings as input and
    uses dot product to score the positive item (inputs["item"]) and also sampled negative
    items (during training).
    Parameters
    ----------
    samplers : List[ItemSampler], optional
        List of item samplers that provide negative samples when `training=True`
    sampling_downscore_false_negatives : bool, optional
        Identify false negatives (sampled item ids equal to the positive item and downscore them
        to the `sampling_downscore_false_negatives_value`), by default True
    sampling_downscore_false_negatives_value : int, optional
        Value to be used to downscore false negatives when
        `sampling_downscore_false_negatives=True`, by default `np.finfo(np.float32).min / 100.0`
    item_id_feature_name: str
        Name of the column containing the item ids
        Defaults to `item_id`
    query_name: str
        Identify query tower for query/user embeddings, by default 'query'
    item_name: str
        Identify item tower for item embeddings, by default'item'
    cache_query: bool
        Add query embeddings to the context block, by default False
    sampled_softmax_mode: bool
        Use sampled softmax for scoring, by default False
    """

    def __init__(
        self,
        samplers: Sequence[ItemSampler] = (),
        sampling_downscore_false_negatives=True,
        sampling_downscore_false_negatives_value: float = MIN_FLOAT,
        item_id_feature_name: str = "item_id",
        query_name: str = "query",
        item_name: str = "item",
        cache_query: bool = False,
        sampled_softmax_mode: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.downscore_false_negatives = sampling_downscore_false_negatives
        self.false_negatives_score = sampling_downscore_false_negatives_value
        self.item_id_feature_name = item_id_feature_name
        self.query_name = query_name
        self.item_name = item_name
        self.cache_query = cache_query

        if not isinstance(samplers, (list, tuple)):
            samplers = (samplers,)  # type: ignore
        self.samplers = samplers
        self.sampled_softmax_mode = sampled_softmax_mode

        self.set_required_features()

    def build(self, input_shapes):
        if isinstance(input_shapes, dict):
            query_shape = input_shapes[self.query_name]
            self.context.add_variable(
                tf.Variable(
                    initial_value=tf.zeros([1, query_shape[-1]], dtype=tf.float32),
                    name="query",
                    trainable=False,
                    validate_shape=False,
                    dtype=tf.float32,
                    shape=tf.TensorShape([None, query_shape[-1]]),
                )
            )
            self.context.add_variable(
                tf.Variable(
                    initial_value=tf.zeros([1, query_shape[-1]], dtype=tf.float32),
                    name="positive_candidates_embeddings",
                    trainable=False,
                    validate_shape=False,
                    dtype=tf.float32,
                    shape=tf.TensorShape([None, query_shape[-1]]),
                )
            )
        super().build(input_shapes)

    def set_required_features(self):
        required_features = set()
        if self.downscore_false_negatives:
            required_features.add(self.item_id_feature_name)

        required_features.update(
            [feature for sampler in self.samplers for feature in sampler.required_features]
        )

        self._required_features = list(required_features)

    def add_features_to_context(self, feature_shapes) -> List[str]:
        return self._required_features

    def _check_input_from_two_tower(self, inputs):
        if set(inputs.keys()) != set([self.query_name, self.item_name]):
            raise ValueError(
                f"Wrong input-names, expected: {[self.query_name, self.item_name]} "
                f"but got: {inputs.keys()}"
            )

    def _check_required_context_item_features_are_present(self):
        not_found = list(
            [
                feat_name
                for feat_name in self._required_features
                if getattr(self, "_context", None) is None
                or feat_name not in self.context.named_variables
            ]
        )

        if len(not_found) > 0:
            raise ValueError(
                f"The following required context features should be available "
                f"for the samplers, but were not found: {not_found}"
            )

    def call(
        self, inputs: Union[tf.Tensor, TabularData], training: bool = True, **kwargs
    ) -> Union[tf.Tensor, TabularData]:
        """Based on the user/query embedding (inputs[self.query_name]), uses dot product to score
            the positive item (inputs["item"]).
            For the sampled-softmax mode, logits are computed by multiplying the query vector
            and the item embeddings matrix (self.context.get_embedding(self.item_column)))
        Parameters
        ----------
        inputs : Union[tf.Tensor, TabularData]
            Dict with the query and item embeddings (e.g. `{"query": <emb>}, "item": <emb>}`),
            where embeddings are 2D tensors (batch size, embedding size)
        training : bool, optional
            Flag that indicates whether in training mode, by default True
        Returns
        -------
        tf.Tensor
            2D Tensor with the scores for the positive items,
            If `training=True`, return the original inputs
        """
        if self.cache_query:
            # enabled only during top-k evaluation
            self.context["query"].assign(tf.cast(inputs[self.query_name], tf.float32))
            self.context["positive_candidates_embeddings"].assign(
                tf.cast(inputs[self.item_name], tf.float32)
            )

        if training:
            return inputs

        if self.sampled_softmax_mode:
            return self._get_logits_for_sampled_softmax(inputs)

        self._check_input_from_two_tower(inputs)
        positive_scores = tf.reduce_sum(
            tf.multiply(inputs[self.query_name], inputs[self.item_name]), keepdims=True, axis=-1
        )
        return positive_scores

    @tf.function
    def call_outputs(
        self, outputs: PredictionOutput, training=True, **kwargs
    ) -> "PredictionOutput":
        """Based on the user/query embedding (inputs[self.query_name]), uses dot product to score
            the positive item and also sampled negative items (during training).
        Parameters
        ----------
        inputs : TabularData
            Dict with the query and item embeddings (e.g. `{"query": <emb>}, "item": <emb>}`),
            where embeddings are 2D tensors (batch size, embedding size)
        training : bool, optional
            Flag that indicates whether in training mode, by default True
        Returns
        -------
        [tf.Tensor,tf.Tensor]
            all_scores: 2D Tensor with the scores for the positive items and, if `training=True`,
            for the negative sampled items too.
            Return tensor is 2D (batch size, 1 + #negatives)
        """
        self._check_required_context_item_features_are_present()
        targets, predictions = outputs.targets, outputs.predictions
        valid_negatives_mask = None

        if self.sampled_softmax_mode or isinstance(targets, tf.Tensor):
            positive_item_ids = targets
        else:
            positive_item_ids = self.context[self.item_id_feature_name]

        if training:
            assert (
                len(self.samplers) > 0
            ), "At least one sampler is required by ItemRetrievalScorer for negative sampling"

            if self.sampled_softmax_mode:
                predictions = self._prepare_query_item_vectors_for_sampled_softmax(
                    predictions, targets
                )

            batch_items_embeddings = predictions[self.item_name]
            batch_items_metadata = self.get_batch_items_metadata()

            positive_scores = tf.reduce_sum(
                tf.multiply(predictions[self.query_name], predictions[self.item_name]),
                keepdims=True,
                axis=-1,
            )

            neg_items_embeddings_list = []
            neg_items_ids_list = []

            # Adds items from the current batch into samplers and sample a number of negatives
            for sampler in self.samplers:
                input_data = EmbeddingWithMetadata(batch_items_embeddings, batch_items_metadata)
                if "item_weights" in sampler._call_fn_args:
                    embedding_table = self.context.get_embedding(self.item_id_feature_name)
                    neg_items = sampler(input_data.__dict__, item_weights=embedding_table)
                else:
                    neg_items = sampler(input_data.__dict__)

                if tf.shape(neg_items.embeddings)[0] > 0:
                    # Accumulates sampled negative items from all samplers
                    neg_items_embeddings_list.append(neg_items.embeddings)
                    if self.downscore_false_negatives:
                        neg_items_ids_list.append(neg_items.metadata[self.item_id_feature_name])
                else:
                    LOG.warn(
                        f"The sampler {type(sampler).__name__} returned no samples for this batch."
                    )

            if len(neg_items_embeddings_list) == 0:
                raise Exception(f"No negative items where sampled from samplers {self.samplers}")
            elif len(neg_items_embeddings_list) == 1:
                neg_items_embeddings = neg_items_embeddings_list[0]
            else:
                neg_items_embeddings = tf.concat(neg_items_embeddings_list, axis=0)

            negative_scores = tf.linalg.matmul(
                predictions[self.query_name], neg_items_embeddings, transpose_b=True
            )

            if self.downscore_false_negatives:
                if isinstance(targets, tf.Tensor):
                    positive_item_ids = targets
                else:
                    positive_item_ids = self.context[self.item_id_feature_name]

                if len(neg_items_ids_list) == 1:
                    neg_items_ids = neg_items_ids_list[0]
                else:
                    neg_items_ids = tf.concat(neg_items_ids_list, axis=0)

                negative_scores, valid_negatives_mask = rescore_false_negatives(
                    positive_item_ids, neg_items_ids, negative_scores, self.false_negatives_score
                )

            predictions = tf.concat([positive_scores, negative_scores], axis=-1)

            # To ensure that the output is always fp32, avoiding numerical
            # instabilities with mixed_float16 policy
            predictions = tf.cast(predictions, tf.float32)

        assert isinstance(predictions, tf.Tensor), "Predictions must be a tensor"
        # prepare targets for computing the loss and metrics
        if self.sampled_softmax_mode and not training:
            # Converts target ids to one-hot representation
            num_classes = tf.shape(predictions)[-1]
            targets_one_hot = tf.one_hot(tf.reshape(targets, (-1,)), num_classes)
            return PredictionOutput(
                predictions,
                targets_one_hot,
                positive_item_ids=positive_item_ids,
                valid_negatives_mask=valid_negatives_mask,
            )
        else:
            # Positives in the first column and negatives in the subsequent columns
            targets = tf.concat(
                [
                    tf.ones([tf.shape(predictions)[0], 1], dtype=predictions.dtype),
                    tf.zeros(
                        [tf.shape(predictions)[0], tf.shape(predictions)[1] - 1],
                        dtype=predictions.dtype,
                    ),
                ],
                axis=1,
            )
            return PredictionOutput(
                predictions,
                targets,
                positive_item_ids=positive_item_ids,
                valid_negatives_mask=valid_negatives_mask,
            )

    def _get_logits_for_sampled_softmax(self, inputs):
        if not isinstance(inputs, tf.Tensor):
            raise ValueError(
                f"Inputs to the Sampled Softmax block should be tensors, got {type(inputs)}"
            )
        embedding_table = self.context.get_embedding(self.item_id_feature_name)
        all_scores = tf.matmul(inputs, tf.transpose(embedding_table))
        return all_scores

    def _prepare_query_item_vectors_for_sampled_softmax(
        self, predictions: tf.Tensor, targets: tf.Tensor
    ):
        # extract positive items embeddings
        if not isinstance(predictions, tf.Tensor):
            raise ValueError(
                f"Inputs to the Sampled Softmax block should be tensors, got {type(predictions)}"
            )
        embedding_table = self.context.get_embedding(self.item_id_feature_name)
        batch_items_embeddings = embedding_ops.embedding_lookup(embedding_table, targets)
        predictions = {self.query_name: predictions, self.item_name: batch_items_embeddings}
        return predictions

    def get_batch_items_metadata(self):
        result = {feat_name: self.context[feat_name] for feat_name in self._required_features}
        return result

    def get_config(self):
        config = super().get_config()
        config = maybe_serialize_keras_objects(self, config, ["samplers"])
        config["downscore_false_negatives"] = self.downscore_false_negatives
        config["false_negatives_score"] = self.false_negatives_score
        config["item_id_feature_name"] = self.item_id_feature_name

        return config

    @classmethod
    def from_config(cls, config):
        config = maybe_deserialize_keras_objects(config, ["samplers"])

        return super().from_config(config)



import pytest
import tensorflow as tf

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.metrics.ranking import AvgPrecisionAt, MRRAt, NDCGAt, PrecisionAt, RecallAt
from merlin.schema import Tags


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_matrix_factorization_model(music_streaming_data: Dataset, run_eagerly, num_epochs=2):
    music_streaming_data.schema = music_streaming_data.schema.remove_by_tag(Tags.TARGET)

    model = mm.MatrixFactorizationModel(music_streaming_data.schema, dim=64)
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    losses = model.fit(music_streaming_data, batch_size=50, epochs=num_epochs)
    assert len(losses.epoch) == num_epochs
    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_two_tower_model(music_streaming_data: Dataset, run_eagerly, num_epochs=2):
    music_streaming_data.schema = music_streaming_data.schema.remove_by_tag(Tags.TARGET)

    model = mm.TwoTowerModel(music_streaming_data.schema, query_tower=mm.MLPBlock([512, 256]))
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    losses = model.fit(music_streaming_data, batch_size=50, epochs=num_epochs)
    assert len(losses.epoch) == num_epochs
    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])


@pytest.mark.parametrize("run_eagerly", [True, False])
@pytest.mark.parametrize("loss", ["categorical_crossentropy", "bpr", "binary_crossentropy"])
def test_two_tower_retrieval_model_with_metrics(ecommerce_data: Dataset, run_eagerly, loss):
    ecommerce_data.schema = ecommerce_data.schema.remove_by_tag(Tags.TARGET)

    metrics = [RecallAt(5), MRRAt(5), NDCGAt(5), AvgPrecisionAt(5), PrecisionAt(5)]
    model = mm.TwoTowerModel(
        schema=ecommerce_data.schema,
        query_tower=mm.MLPBlock([128, 64]),
        samplers=[mm.InBatchSampler()],
        metrics=metrics,
        loss=loss,
    )
    # Setting up evaluation
    model.set_retrieval_candidates_for_evaluation(ecommerce_data)
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    # Training
    num_epochs = 2
    losses = model.fit(
        ecommerce_data,
        batch_size=10,
        epochs=num_epochs,
        train_metrics_steps=3,
        validation_data=ecommerce_data,
        validation_steps=3,
    )
    assert len(losses.epoch) == num_epochs

    # Checking train metrics
    expected_metrics = ["recall_at_5", "mrr_at_5", "ndcg_5", "map_at_5", "precision_at_5"]
    expected_loss_metrics = ["loss", "regularization_loss", "total_loss"]
    expected_metrics_all = expected_metrics + expected_loss_metrics
    assert len(expected_metrics_all) == len(
        set(losses.history.keys()).intersection(set(expected_metrics_all))
    )
    for metric_name in expected_metrics + expected_loss_metrics:
        assert len(losses.history[metric_name]) == num_epochs
        if metric_name in expected_metrics:
            assert losses.history[metric_name][1] >= losses.history[metric_name][0]
        elif metric_name in expected_loss_metrics:
            assert losses.history[metric_name][1] <= losses.history[metric_name][0]

    _ = model.evaluate(ecommerce_data, batch_size=10)


def test_retrieval_evaluation_without_negatives(ecommerce_data: Dataset):
    model = mm.TwoTowerModel(schema=ecommerce_data.schema, query_tower=mm.MLPBlock([64]))
    model.compile(optimizer="adam", run_eagerly=True)
    model.fit(ecommerce_data, batch_size=50)
    with pytest.raises(ValueError) as exc_info:
        model.evaluate(ecommerce_data, batch_size=10)
        assert "You need to specify the set of negatives to use for evaluation" in str(
            exc_info.value
        )


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_youtube_dnn_retrieval(
    sequence_testing_data: Dataset,
    run_eagerly: bool,
):
    model = mm.YoutubeDNNRetrievalModel(schema=sequence_testing_data.schema, max_seq_length=4)
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    losses = model.fit(sequence_testing_data, batch_size=50, epochs=2)

    assert len(losses.epoch) == 2
    for metric in losses.history.keys():
        assert type(losses.history[metric]) is list
    batch = mm.sample_batch(
        sequence_testing_data, batch_size=10, include_targets=False, to_dense=True
    )
    out = model({k: tf.cast(v, tf.int64) for k, v in batch.items()})

    assert out.shape[-1] == 51997



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Sequence, Type, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer

from merlin.models.config.schema import SchemaMixin
from merlin.models.tf.typing import TabularData
from merlin.models.tf.utils.mixins import ModelLikeBlock
from merlin.models.utils.registry import Registry
from merlin.schema import Schema, Tags

block_registry: Registry = Registry.class_registry("tf.blocks")
BlockType = Union["Block", str, Sequence[str]]


if TYPE_CHECKING:
    from merlin.models.tf.blocks.core.combinators import (
        Filter,
        ParallelBlock,
        SequentialBlock,
        TabularAggregationType,
    )
    from merlin.models.tf.models.base import Model, RetrievalModel
    from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask


class PredictionOutput(NamedTuple):
    predictions: Union[TabularData, tf.Tensor]
    targets: Union[TabularData, tf.Tensor]
    positive_item_ids: Optional[tf.Tensor] = None
    label_relevant_counts: Optional[tf.Tensor] = None
    valid_negatives_mask: Optional[tf.Tensor] = None

    def copy_with_updates(
        self,
        predictions: Optional[Union[TabularData, tf.Tensor]] = None,
        targets: Optional[Union[TabularData, tf.Tensor]] = None,
        positive_item_ids: Optional[tf.Tensor] = None,
        label_relevant_counts: Optional[tf.Tensor] = None,
        valid_negatives_mask: Optional[tf.Tensor] = None,
    ):
        """Creates a new instance of PredictionOutput
        allowing to override the attributes for the copy
        """
        output = PredictionOutput(
            predictions=(self.predictions if predictions is None else predictions),
            targets=(self.targets if targets is None else targets),
            positive_item_ids=(
                self.positive_item_ids if positive_item_ids is None else positive_item_ids
            ),
            label_relevant_counts=(
                self.label_relevant_counts
                if label_relevant_counts is None
                else label_relevant_counts
            ),
            valid_negatives_mask=(
                self.valid_negatives_mask if valid_negatives_mask is None else valid_negatives_mask
            ),
        )
        return output


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class BlockContext(Layer):
    """BlockContext is part of each block.

    It is used to store/retrieve public variables, and can be used to retrieve features.
    This is created automatically in the model and doesn't need to be created manually.

    """

    def __init__(self, **kwargs):
        feature_names = kwargs.pop("feature_names", [])
        feature_dtypes = kwargs.pop("feature_dtypes", {})
        super(BlockContext, self).__init__(**kwargs)
        self._feature_names = feature_names
        self._feature_dtypes = feature_dtypes

    def add_embedding_weight(self, name, **kwargs):
        table = self.add_weight(name=f"{str(name)}/embedding", **kwargs)

        return table

    def add_features(self, *name):
        self._feature_names = list({*self._feature_names, *name})

    def add_variable(self, variable):
        setattr(self, variable.name, variable)

    def set_dtypes(self, features):
        for feature_name in features:
            feature = features[feature_name]

            if isinstance(feature, tuple):
                dtype = feature[0].dtype
            else:
                dtype = feature.dtype

            self._feature_dtypes[feature_name] = dtype

    def __getitem__(self, item):
        if isinstance(item, Schema):
            if len(item.column_names) > 1:
                raise ValueError("Schema contains more than one column.")
            item = item.column_names[0]
        elif isinstance(item, Tags):
            item = item.value
        else:
            item = str(item)
        return self.named_variables[item]

    def get_embedding(self, item):
        if isinstance(item, Tags):
            item = item.value
        else:
            item = str(item)
        return self.named_variables[f"{item}/embedding"]

    def get_mask(self):
        mask_schema = self.named_variables.get("masking_schema", None)
        if mask_schema is None:
            raise ValueError(
                "The mask schema is not stored, " "please make sure that a MaskingBlock was set"
            )
        return mask_schema

    @property
    def named_variables(self) -> Dict[str, tf.Variable]:
        outputs = {}
        for var in self.variables:
            if var.name.endswith("/embedding:0"):
                name = "/".join(var.name.split("/")[-2:])
            else:
                name = var.name.split("/")[-1]
            outputs[name.replace(":0", "")] = var

        return outputs

    def _merge(self, other: "BlockContext"):
        self.public_variables.update(other.public_variables)
        self._feature_names = list(set(self._feature_names + other._feature_names))

    def build(self, input_shape):
        for feature_name in self._feature_names:
            if feature_name not in self.named_variables:
                shape = input_shape[feature_name]
                dtype = self._feature_dtypes.get(feature_name, tf.float32)

                if len(tuple(shape)) == 2:
                    s = (
                        shape[-1]
                        if not isinstance(shape[-1], (tuple, tf.TensorShape))
                        else int(shape[0][0] / shape[-1][0])
                    )
                    if s == 1:
                        shape = tf.TensorShape(
                            [
                                None,
                            ]
                        )
                        var = tf.zeros([1], dtype=dtype)
                    else:
                        shape = tf.TensorShape([None, s])
                        var = tf.zeros([1, s], dtype=dtype)
                elif tuple(shape) != (None,):
                    var = tf.zeros((shape), dtype=dtype)
                else:
                    var = tf.zeros([1], dtype=dtype)

                setattr(
                    self,
                    feature_name,
                    tf.Variable(
                        var,
                        name=feature_name,
                        trainable=False,
                        dtype=dtype,
                        shape=shape,
                    ),
                )

        super(BlockContext, self).build(input_shape)

    def call(self, features, **kwargs):
        for feature_name in self._feature_names:
            self.named_variables[feature_name].assign(features[feature_name])

        return features

    def get_config(self):
        config = super(BlockContext, self).get_config()
        config["feature_names"] = self._feature_names
        config["feature_dtypes"] = self._feature_dtypes

        return config


class ContextMixin:
    @property
    def context(self) -> BlockContext:
        return self._context

    def _set_context(self, context: BlockContext):
        if hasattr(self, "_context"):
            context._merge(self._context)
        self._context: BlockContext = context


class Block(SchemaMixin, ContextMixin, Layer):
    """Core abstraction in Merlin models."""

    registry = block_registry

    def __init__(self, context: Optional[BlockContext] = None, **kwargs):
        super(Block, self).__init__(**kwargs)
        if context:
            self._set_context(context)

    @classmethod
    @tf.autograph.experimental.do_not_convert
    def parse(cls, *block: BlockType) -> "Block":
        if len(block) == 1 and isinstance(block[0], (list, tuple)):
            block = block[0]  # type: ignore

        output: "Block"
        if len(block) == 1:
            output = cls.registry.parse(block[0])
        else:
            blocks = [cls.registry.parse(b) for b in block]
            output = blocks[0].connect(*blocks[1:])

        return output

    @classmethod
    def from_layer(cls, layer: tf.keras.layers.Layer) -> "Block":
        layer.__class__ = cls

        return layer  # type: ignore

    @classmethod
    def parse_block(cls, input: Union["Block", tf.keras.layers.Layer]) -> "Block":
        if isinstance(input, Block):
            return input

        return cls.from_layer(input)

    def build(self, input_shapes):
        self._maybe_propagate_context(input_shapes)

        return super().build(input_shapes)

    def _maybe_build(self, inputs):
        if getattr(self, "_context", None) and not self.context.built:
            self.context.set_dtypes(inputs)

        super()._maybe_build(inputs)

    def call_outputs(
        self, outputs: PredictionOutput, training=False, **kwargs
    ) -> "PredictionOutput":
        return outputs

    def register_features(self, feature_shapes) -> List[str]:
        return []

    def as_tabular(self, name=None) -> "Block":
        from merlin.models.tf.blocks.core.combinators import SequentialBlock
        from merlin.models.tf.blocks.core.tabular import AsTabular

        if not name:
            name = self.name

        return SequentialBlock([self, AsTabular(name)], copy_layers=False)

    def repeat(self, num: int = 1) -> "SequentialBlock":
        """Repeat the block num times.

        Parameters
        ----------
        num : int
            Number of times to repeat the block.
        """
        from merlin.models.tf.blocks.core.combinators import SequentialBlock

        repeated = []
        for _ in range(num):
            repeated.append(self.copy())

        return SequentialBlock(repeated)

    def prepare(
        self,
        block: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
    ) -> "SequentialBlock":
        """Transform the inputs of this block.

        Parameters
        ----------
        block: Optional[Block]
            If set, this block will be used to transform the inputs of this block.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the inputs.

        """
        from merlin.models.tf.blocks.core.combinators import SequentialBlock, TabularBlock

        block = TabularBlock(post=post, aggregation=aggregation) or block

        return SequentialBlock([block, self])

    def repeat_in_parallel(
        self,
        num: int = 1,
        prefix=None,
        names: Optional[List[str]] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
        copies=True,
        residual=False,
        **kwargs,
    ) -> "ParallelBlock":
        """Repeat the block num times in parallel.

        Parameters
        ----------
        num: int
            Number of times to repeat the block.
        prefix: str
            Prefix to use for the names of the blocks.
        names: List[str]
            Names of the blocks.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the inputs.
        copies: bool
            Whether to copy the block or not.
        residual: bool
            Whether to use a residual connection or not.

        """
        from merlin.models.tf.blocks.core.base import NoOp
        from merlin.models.tf.blocks.core.combinators import ParallelBlock

        repeated = {}
        iterator = names if names else range(num)
        if not names and prefix:
            iterator = [f"{prefix}{num}" for num in iterator]
        for name in iterator:
            repeated[str(name)] = self.copy() if copies else self

        if residual:
            repeated["shortcut"] = NoOp()

        return ParallelBlock(repeated, post=post, aggregation=aggregation, **kwargs)

    def connect(
        self,
        *block: Union[tf.keras.layers.Layer, str],
        block_name: Optional[str] = None,
        context: Optional[BlockContext] = None,
    ) -> Union["SequentialBlock", "Model", "RetrievalModel"]:
        """Connect the block to other blocks sequentially.

        Parameters
        ----------
        block: Union[tf.keras.layers.Layer, str]
            Blocks to connect to.
        block_name: str
            Name of the block.
        context: Optional[BlockContext]
            Context to use for the block.

        """
        from merlin.models.tf.blocks.core.combinators import SequentialBlock
        from merlin.models.tf.models.base import Model, RetrievalBlock, RetrievalModel
        from merlin.models.tf.prediction_tasks.retrieval import ItemRetrievalTask

        blocks = [self.parse(b) for b in block]

        for b in blocks:
            if isinstance(b, Block):
                if not b.has_schema and self.has_schema:
                    b._schema = self.schema

        output = SequentialBlock(
            [self, *blocks], copy_layers=False, block_name=block_name, context=context
        )

        if isinstance(blocks[-1], ModelLikeBlock):
            if (
                any(isinstance(b, RetrievalBlock) for b in blocks)
                or isinstance(self, RetrievalBlock)
                and any(isinstance(b, ItemRetrievalTask) for b in blocks)
            ):
                return RetrievalModel(output)

            return Model(output)

        return output

    def connect_with_residual(
        self,
        block: Union[tf.keras.layers.Layer, str],
        activation=None,
    ) -> "SequentialBlock":
        """Connect the block to other blocks sequentially with a residual connection.

        Parameters
        ----------
        block: Union[tf.keras.layers.Layer, str]
            Blocks to connect to.
        activation: str
            Activation to use for the residual connection.

        """
        from merlin.models.tf.blocks.core.combinators import ResidualBlock, SequentialBlock

        _block = self.parse(block)
        residual_block = ResidualBlock(_block, activation=activation)

        if isinstance(self, SequentialBlock):
            self.layers.append(residual_block)

            return self

        return SequentialBlock([self, residual_block], copy_layers=False)

    def connect_with_shortcut(
        self,
        block: Union[tf.keras.layers.Layer, str],
        shortcut_filter: Optional["Filter"] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
        block_outputs_name: Optional[str] = None,
    ) -> "SequentialBlock":
        """Connect the block to other blocks sequentially with a shortcut connection.

        Parameters
        ----------
        block: Union[tf.keras.layers.Layer, str]
            Blocks to connect to.
        shortcut_filter: Filter
            Filter to use for the shortcut connection.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the outputs.
        block_outputs_name: str
            Name of the block outputs.
        """
        from merlin.models.tf.blocks.core.combinators import SequentialBlock, WithShortcut

        _block = self.parse(block) if not isinstance(block, Block) else block
        residual_block = WithShortcut(
            _block,
            shortcut_filter=shortcut_filter,
            post=post,
            aggregation=aggregation,
            block_outputs_name=block_outputs_name,
        )

        if isinstance(self, SequentialBlock):
            self.layers.append(residual_block)

            return self

        return SequentialBlock([self, residual_block], copy_layers=False)

    def connect_debug_block(self, append=True):
        """Connect the block to a debug block.

        Parameters
        ----------
        append: bool
            Whether to append the debug block to the block or to prepend it.
        """
        from merlin.models.tf.blocks.core.base import Debug
        from merlin.models.tf.blocks.core.combinators import SequentialBlock

        if not append:
            return SequentialBlock([Debug(), self])

        return self.connect(Debug())

    def connect_branch(
        self,
        *branches: Union["Block", "PredictionTask", str],
        add_rest=False,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
        **kwargs,
    ) -> Union["SequentialBlock", "Model", "RetrievalModel"]:
        """Connect the block to one or multiple branches.

        Parameters
        ----------
        branches: Union[Block, PredictionTask, str]
            Blocks to connect to.
        add_rest: bool
            Whether to add the rest of the block to the branches.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the outputs.

        """
        from merlin.models.tf.blocks.core.combinators import Filter, ParallelBlock, SequentialBlock

        _branches = [self.parse(b) for b in branches]

        all_features = []
        for branch in _branches:
            if getattr(branch, "set_schema", None):
                branch.set_schema(self.schema)
            if isinstance(branch, SequentialBlock):
                filter_features = branch.filter_features
                if filter_features:
                    all_features.extend(filter_features)

        if add_rest:
            if not self.schema:
                raise ValueError("Schema is required to add rest features.")
            rest_features = self.schema.without(list(set([str(f) for f in all_features])))
            rest_block = SequentialBlock([Filter(rest_features)])
            _branches.append(rest_block)

        if all(isinstance(branch, ModelLikeBlock) for branch in _branches):
            from merlin.models.tf import ParallelPredictionBlock

            parallel = ParallelPredictionBlock(
                *_branches, post=post, aggregation=aggregation, **kwargs
            )

            return Model(SequentialBlock([self, parallel]))

        return SequentialBlock(
            [self, ParallelBlock(*_branches, post=post, aggregation=aggregation, **kwargs)]
        )

    def to_model(
        self,
        schema: Schema,
        input_block: Optional[Block] = None,
        prediction_tasks: Optional[
            Union["PredictionTask", List["PredictionTask"], "ParallelPredictionBlock"]
        ] = None,
        **kwargs,
    ) -> "Model":
        """Wrap the block between inputs & outputs to create a model.

        Parameters
        ----------
        schema: Schema
            Schema to use for the model.
        input_block: Optional[Block]
            Block to use as input.
        prediction_tasks: Optional[
            Union[PredictionTask, List[PredictionTask], ParallelPredictionBlock]
        ]
            Prediction tasks to use.

        """
        from merlin.models.tf.blocks.core.inputs import InputBlock
        from merlin.models.tf.models.base import Model
        from merlin.models.tf.models.utils import parse_prediction_tasks

        if is_input_block(self.first):
            if input_block is not None:
                raise ValueError("The block already includes an InputBlock")
            input_block = self.first

        aggregation = kwargs.pop("aggregation", "concat")
        _input_block: Block = input_block or InputBlock(schema, aggregation=aggregation, **kwargs)

        prediction_tasks = parse_prediction_tasks(schema, prediction_tasks)

        return Model(_input_block, self, prediction_tasks)

    def select_by_name(self, name: str) -> Optional["Block"]:
        if name == self.name:
            return self

        return None

    def copy(self):
        return self.from_config(self.get_config())

    def _maybe_propagate_context(self, input_shapes):
        if getattr(self, "_context", None) and not self.context.built:
            for module in self.submodules:
                if hasattr(module, "_set_context") and not getattr(module, "context", False):
                    module._set_context(self.context)
                if hasattr(module, "add_features_to_context") and not getattr(
                    module, "_features_registered", False
                ):
                    feature_names = module.add_features_to_context(input_shapes)
                    module._features_registered = True
                    if feature_names:
                        self.context.add_features(*feature_names)
            self._need_to_call_context = True
            self.context.build(input_shapes)

    def __rrshift__(self, other):
        return right_shift_layer(self, other)


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class NoOp(tf.keras.layers.Layer):
    def call(self, inputs, **kwargs):
        return inputs

    def compute_output_shape(self, input_shape):
        return input_shape


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class Debug(tf.keras.layers.Layer):
    def call(self, inputs, **kwargs):
        return inputs

    def compute_output_shape(self, input_shape):
        return input_shape


def name_fn(name, inp):
    return "/".join([name, inp]) if name else None


MetricOrMetricClass = Union[tf.keras.metrics.Metric, Type[tf.keras.metrics.Metric]]
MetricOrMetrics = Union[Sequence[MetricOrMetricClass], MetricOrMetricClass]


@dataclass
class EmbeddingWithMetadata:
    embeddings: tf.Tensor
    metadata: Dict[str, tf.Tensor]


def is_input_block(block: Block) -> bool:
    is_defined = True if block else False
    return is_defined and getattr(block, "is_input", False)


def has_input_block(block: Block) -> bool:
    from merlin.models.tf.blocks.core.combinators import SequentialBlock

    if isinstance(block, SequentialBlock):
        return block.inputs is not None and is_input_block(block.inputs)
    return is_input_block(block.inputs)


def _output_metrics(metrics):
    if len(metrics) == 1:
        return metrics[list(metrics.keys())[0]]

    return metrics


def right_shift_layer(self, other):
    from merlin.models.tf.blocks.core.combinators import Filter, SequentialBlock

    if isinstance(other, (list, Tags)):
        left_side = [Filter(other)]
    else:
        left_side = other.layers if isinstance(other, SequentialBlock) else [other]
    right_side = self.layers if isinstance(self, SequentialBlock) else [self]

    return SequentialBlock(left_side + right_side)
